# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._settings import settings
from ..utils import colors, style_axis
from chemicals.identifiers import to_searchable_format
from .bubble_point import BubblePoint
from .lle import LLE
from ..indexer import MaterialIndexer
import matplotlib.pyplot as plt
from math import floor
import thermosteam as tmo
import numpy as np
from numpy.linalg import lstsq
import flexsolve as flx
from scipy import interpolate
from scipy.ndimage.filters import gaussian_filter
from typing import Iterable

__all__ = (
    'plot_vle_binary_phase_envelope',
    'plot_vle_phase_envelope',
    'plot_lle_ternary_diagram',
    'ternary_composition_grid',
)


# %% Utilities

def ternary_composition_grid(N, grid=None): # pragma: no cover
    grid = grid or []
    for z in np.linspace(0, 1 - 1/N, N):
        xs = np.linspace(0, 1-z, N)
        ys = 1 - z - xs 
        zs = z * np.ones(N)
        mesh = np.array([xs, ys, zs])
        grid.append(mesh)
    mesh = np.array([0, 0, 1]).reshape([3, 1])
    grid.append(mesh)
    return np.hstack(grid).transpose()

def as_thermo(thermo, chemicals): # pragma: no cover
    try:
        thermo = settings.get_default_thermo(thermo)
    except:
        thermo = tmo.Thermo(chemicals)
    else:
        thermo = tmo.Thermo(chemicals,
                            Gamma=thermo.Gamma,
                            Phi=thermo.Phi,
                            PCF=thermo.PCF)
    return thermo

# %% Plot functions

def plot_vle_binary_phase_envelope(chemicals, T=None, P=None, vc=None, lc=None, thermo=None,
                                   yticks=None, line_styles=None, N=None): # pragma: no cover
    """
    Plot the binary phase envelope of two chemicals at a given temperature or pressure.

    Parameters
    ----------
    chemicals : Iterable[Chemical or str]
        Chemicals in equilibrium.
    T : float|list[float], optional
        Temperature [K]. 
    P : float|list[float], optional
        Pressure [Pa]. 
    vc : str, optional
        Color of vapor line.
    lc : str, optional
        Color of liquid line.
    thermo : Thermo, optional
        Thermodynamic property package.

    Examples
    --------
    >>> # from thermosteam import equilibrium as eq
    >>> # eq.plot_vle_binary_phase_envelope(['Ethanol', 'Water'], P=101325)
    
    .. figure:: ../images/water_ethanol_binary_phase_envelope.png

    """
    if line_styles is None: line_styles = ['-', '-.', '--']
    thermo = as_thermo(thermo, chemicals)
    chemical_a, chemical_b = chemicals = [thermo.as_chemical(i) for i in chemicals]
    BP = tmo.BubblePoint(chemicals, thermo)
    if N is None: N = 50
    zs_a = np.linspace(0, 1, N)
    zs_b = 1 - zs_a
    zs = np.vstack([zs_a, zs_b]).transpose()
    if P:
        if not isinstance(P, Iterable): P = [P]
        Xs = P
        X_units = 'Pa'
        assert not T, "must pass either T or P, but not both"
        bpss = [[BP(z, P=i) for z in zs] for i in P]
        mss = [[bp.T for bp in bps] for bps in bpss]
        ylabel = 'Temperature [K]'
    elif T:
        if not isinstance(T, Iterable): T = [T]
        assert not P, "must pass either T or P, but not both"
        Xs = T
        X_units = 'K'
        bpss = [[BP(z, T=i) for z in zs] for i in T]
        mss = [[bp.P for bp in bps] for bps in bpss]
        ylabel = 'Pressure [Pa]'
    else:
        raise AssertionError("must pass either T or P")
    plt.figure()
    mss = np.array(mss)
    for X, ms, bps, ls in zip(Xs, mss, bpss, line_styles):
        ys_a = np.array([bp.y[0] for bp in bps])
        ms_liq = ms.copy()
        ms_gas = interpolate.interp1d(ys_a, ms, bounds_error=False, kind='slinear')(zs_a)
        # if P:
        #     azeotrope = ms_liq > ms_gas
        # elif T:
        #     azeotrope = ms_liq < ms_gas
        # ms_liq[0] = 0.5 * (ms_liq[0] + ms_gas[0])
        # ms_liq[-1] = 0.5 * (ms_liq[-1] + ms_gas[-1])
        # if azeotrope.any():
        #     index = np.where(azeotrope)[0]
        #     left = index[0]
        #     right = index[-1]
        #     mid = int(np.median(index))
        #     azeotrope[left:right] = True
        #     azeotrope[mid] = False
        #     ms_gas[mid] = ms_liq[mid]
        #     ms_gas = interpolate.interp1d(zs_a[~azeotrope], ms_gas[~azeotrope], bounds_error=False, kind='slinear')(zs_a)
        plt.plot(zs_a, ms_gas, c=vc if vc is not None else colors.red.RGBn, label=f'vapor [{int(X)} {X_units}]', ls=ls)
        plt.plot(zs_a, ms_liq, c=lc if lc is not None else colors.blue.RGBn, label=f'liquid [{int(X)} {X_units}]', ls=ls)
    plt.xlim([0, 1])
    plt.ylim([mss.min(), mss.max()])
    if yticks is None: yticks, ytext = plt.yticks()
    plt.legend()
    plt.xlabel(f'{chemical_a} molar fraction')
    plt.ylabel(ylabel)
    style_axis(xticks=np.linspace(0, 1, 5), yticks=yticks)
    
def plot_vle_phase_envelope(
        IDs, zs, T_range=None, P_range=None, thermo=None, vc=None, lc=None, 
        xticks=None, yticks=None, N=None, line_styles=None, labels=None): # pragma: no cover
    """
    Plot the binary phase envelope of two chemicals at a given temperature or pressure.

    Parameters
    ----------
    IDs : Iterable[str]
        IDs of chemicals in equilibrium.
    zs : Iterable[float]
        Molar fraction of chemicals in equilibrium.
    T_range : tuple[float, float] optional
        Temperature [K]. 
    P_range : tuple[float, float] optional
        Pressure [atm]. 
    thermo : Thermo|Iterable[Thermo], optional
        Thermodynamic property package(s).
    vc : str, optional
        Color of vapor line.
    lc : str, optional
        Color of liquid line.

    """
    if line_styles is None: line_styles = ['-', '-.', '--']
    if thermo is None: 
        thermo = [tmo.settings.get_thermo()]
    elif isinstance(thermo, (tmo.Thermo, tmo.IdealThermo)):
        thermo = [thermo]
    if labels is None:
        N_thermo = len(thermo)
        if N_thermo == 1:
            labels = ['']
        else:
            labels = [str(i) for i in range(N_thermo)]
    BPs = [tmo.BubblePoint(t.chemicals[IDs], t) for t in thermo]
    DPs = [tmo.DewPoint(t.chemicals[IDs], t) for t in thermo]
    if N is None: N = 20
    if P_range is not None:
        assert T_range is None, "must pass either T_range or P_range, but not both"
        xlim = P_range
        xs = np.linspace(*xlim, N)
        bpss = [
            [BP(zs, P=P * 101325) for P in xs]
            for BP in BPs
        ]
        dpss = [
            [DP(zs, P=P * 101325) for P in xs]
            for DP in DPs
        ]
        ylabel = 'Temperature [K]'
        xlabel = 'Pressure [atm]'
        variable = 'T'
    elif T_range:
        assert P_range is None, "must pass either T_range or P_range, but not both"
        xlim = T_range
        xs = np.linspace(*xlim, N)
        bpss = [
            [BP(zs, T=T) for T in xs]
            for BP in BPs
        ]
        dpss = [
            [DP(zs, T=T) for T in xs]
            for DP in DPs
        ]
        ylabel = 'Pressure [atm]'
        xlabel = 'Temperature [K]'
        variable = 'P'
    else:
        raise AssertionError("must pass either T or P")
    Lss = np.array([
        [getattr(bp, variable) for bp in bps]
        for bps in bpss
    ])
    Vss = np.array([
        [getattr(dp, variable) for dp in dps]
        for dps in dpss
    ])
    plt.figure()
    if vc is None: vc = colors.red.RGBn
    if lc is None: lc = colors.blue.RGBn
    for Vs, Ls, ls, label in zip(Vss, Lss, line_styles, labels):
        plt.plot(xs, Vs, c=vc, ls=ls, label=', '.join([label, 'vapor']))
        plt.plot(xs, Ls, c=lc, ls=ls, label=', '.join([label, 'liquid']))
    plt.xlim(xlim)
    plt.ylim([min(Lss.min(), Vss.min()), max(Lss.max(), Vss.max())])
    plt.legend()
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    style_axis(xticks=xticks, yticks=yticks)
    
def plot_lle_ternary_diagram(
        carrier, solvent, solute, T, P=101325, thermo=None, color=None,
        tie_line_points=None,
        tie_color=None,
        N_tie_lines=8,
        N_equilibrium_grids=8,
        method=None
    ): # pragma: no cover
    """
    Plot the ternary phase diagram of chemicals in liquid-liquid equilibrium.

    Parameters
    ----------
    carrier : Chemical
    solvent : Chemical
    solute : Chemical
    T : float, optional
        Temperature [K]. 
    P : float, optional
        Pressure [Pa]. Defaults to 101325.
    thermo : Thermo, optional
        Thermodynamic property package.
    color : str, optional
        Color of equilibrium line.
    tie_line_points : 1d array(size=3), optional
        Additional composition points to create tie lines.
    tie_color : str, optional
        Color of tie lines.
    N_tie_lines : int, optional
        Number of tie lines. The default is 15.
    N_equilibrium_grids : int, optional
        Number of solute composition points to plot. The default is 15.

    Examples
    --------
    >>> # from thermosteam import equilibrium as eq
    >>> # eq.plot_lle_ternary_diagram('Water', 'Ethanol', 'EthylAcetate', T=298.15)
    
    .. figure:: ../images/water_ethanol_ethyl_acetate_ternary_diagram.png

    """
    import ternary
    chemicals = [carrier, solute, solvent]
    thermo = as_thermo(thermo, chemicals)
    chemicals = thermo.chemicals
    IDs = chemicals.IDs
    imol = MaterialIndexer(['l', 'L'], chemicals=chemicals)
    data = imol.data
    lle = LLE(imol, thermo=thermo, method=method)
    composition_grid = ternary_composition_grid(N_equilibrium_grids, tie_line_points)
    tie_lines = []
    MW = chemicals.MW
    for zs in composition_grid:
        data[0] = zs
        data[1] = 0
        lle(T, P)
        mass = data * MW
        lL = mass.sum(1, keepdims=True) 
        if (abs(lL) > 1e-5).all():
            tie_line = 100 * mass / lL
            tie_diff = np.abs(tie_line[0] - tie_line[1]) ** 0.75
            if tie_diff.sum() > 0.2:
                tie_lines.append(tie_line)
    tie_points = np.vstack(tie_lines)
    tie_points = tie_points[np.argsort(tie_points[:, 0])]
    
    def f(x):
        shape = np.shape(x)
        X = np.ones([*shape, 5])
        X[..., 0] = x2 = x * x
        X[..., 1] = np.sqrt(x)
        X[..., 2] = x
        X[..., 3] = x2 * x
        return X
    
    xs = tie_points[:, 0]
    ys = tie_points[:, 1]
    X = f(xs)
    result = lstsq(X, ys, rcond=None)
    coef = result[0]
    xs = np.linspace(0, 100, 500)
    ys = (f(xs) * coef).sum(axis=1)
    zs = 100 - xs - ys
    ys[ys < 0] = 0
    zs[zs < 0] = 0
    tie_points = np.zeros([len(xs), 3])
    tie_points[:, 0] = xs
    tie_points[:, 1] = ys
    tie_points[:, 2] = zs
    tie_points /= tie_points.sum(axis=1, keepdims=True) / 100
    fig, tax = ternary.figure(scale=100)
    label = "-".join(IDs) + " Phase Diagram"
    tax.boundary(linewidth=2.0)
    tax.gridlines(color=colors.grey_tint.RGBn, multiple=10)
    fontsize = 12
    offset = 0.14
    C, A, S = [to_searchable_format(i) for i in IDs]
    tax.right_corner_label(8*' ' + 'Solvent', fontsize=fontsize)
    tax.top_corner_label('Solute\n', fontsize=fontsize)
    tax.left_corner_label('Carrier' + 8*' ', fontsize=fontsize)
    tax.left_axis_label(f'{C} wt. %', fontsize=fontsize,
                        offset=offset)
    tax.right_axis_label(f'{A} wt. %', fontsize=fontsize,
                         offset=offset)
    tax.bottom_axis_label(f'{S} wt. %', fontsize=fontsize,
                          offset=offset)
    if color is None: color = 'k'
    tax.plot(tie_points, c=colors.neutral_shade.RGBn, label=label)
    
    if tie_line_points is None:
        assert N_tie_lines, "must specify number of tie lines if no equilibrium points are given"
        xs, ys = zip(*tie_lines)
        new_xs = np.array(xs)
        new_ys = np.array(ys)
        # new_xs = []
        # new_ys = []
        # for left, right in zip(xs, ys):
        #     x, *_ = left
        #     y = (f(x) * coef).sum()
        #     z = 100 - y - x
        #     if y < 0 or z < 0: continue
        #     left = np.array([x, y, z])
        #     left = 100 * left / left.sum()
        #     x, *_ = right
        #     y = (f(x) * coef).sum()
        #     z = 100 - y - x
        #     if y < 0 or z < 0: continue
        #     right = np.array([x, y, z])
        #     right = 100 * right / right.sum()
        #     new_xs.append(left)
        #     new_ys.append(right)
        # new_xs = np.array(new_xs)
        # new_ys = np.array(new_ys)
        # index = np.argsort(new_xs[:, 0])
        N_total_lines = len(new_xs) 
        step = floor(N_total_lines / N_tie_lines) or 1
        partition_index = np.arange(0, N_total_lines, step, dtype=int)
        # index = index[partition_index]
        new_xs = [new_xs[i] for i in partition_index]
        new_ys = [new_ys[i] for i in partition_index]
        tie_lines = list(zip(new_xs, new_ys))    
    else:
        assert N_tie_lines is None, "cannot specify number of tie lines if equilibrium points are given"
        N_tie_lines = len(tie_line_points)
        tie_lines = tie_lines[:N_tie_lines]
    if tie_color is None: tie_color = colors.grey_shade.RGBn
    for x, y in tie_lines:
        tax.line(x, y, ls='--', c=tie_color)
    
    tax.ticks(axis='lbr', multiple=10, linewidth=1, offset=0.025)
    tax.get_axes().axis('off')
    tax.clear_matplotlib_ticks()


# -*- coding: utf-8 -*-
"""
Created on Fri Sep 26 17:14:27 2025

@author: yoelr
"""
from inspect import signature
from numba import njit
from numba.types import Tuple

__all__ = ('JitSignature',)

class JitSignature:
    __slots__ = ('types',)
    def __init__(self, **types):
        self.types = types
        
    def __call__(self, output):
        types = self.types
        outputs = output.split(', ')
        if len(outputs) == 1:
            output = types[output]
        else:
            output = Tuple([types[i] for i in outputs])
        return lambda f: njit(output(*[types[i] for i in signature(f).parameters]), cache=True)(f)
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
# DO NOT DELETE:
# chemicals' numba feature does not offer any benefit to speed
# but it may be used later
# def use_numba_chemicals():
#     import chemicals
#     import chemicals.numba as numba
#     Module = type(chemicals)
#     try: numba.Lastovka_Shaw # First autoload with __getattr__
#     except: pass
#     def update_module(module, other,
#             ignored=chemicals.utils.numba_cache_blacklisted,
#         ):
#         dct = module.__dict__
#         isa = isinstance
#         isfunc = callable
#         for i, j in other.__dict__.items():
#             if i not in dct or i in ignored: continue
#             if isa(j, Module) and 'chemicals.' in j.__name__:
#                 update_module(dct[i], j)
#             else:
#                 if isfunc(j) and 'numba' in repr(j.__class__): dct[i] = j
#     update_module(chemicals, numba)
# use_numba_chemicals()
# del use_numba_chemicals
__version__ = "0.52.16"

from . import thermo
del thermo
from . import (
    constants,
    base,
    utils,
    chemicals,
    exceptions,
    functional,
    units_of_measure,
    separations,
    functors,
    phenomenode,
)
from .utils import docround
from .network import *
from ._chemical_data import ChemicalData
from ._chemical import Chemical
from ._chemicals import Chemicals, CompiledChemicals, ChemicalDraft, ChemicalsOutline
from ._thermal_condition import ThermalCondition
from ._thermo import Thermo, IdealThermo
from ._settings import settings, ProcessSettings
from ._thermo_data import ThermoData
from . import (
    indexer,
    reaction,
    equilibrium,
    mixture,
    network,
)
from ._stream import Stream
from ._heat_and_power import Heat, Power
from ._multi_stream import MultiStream
from .base import functor
from .reaction import *
from .equilibrium import * 
from .mixture import *
from ._preferences import preferences
from .phenomenode import *

__all__ = ('Chemical', 'ChemicalData', 'Chemicals', 'CompiledChemicals', 
           'ChemicalDraft', 'ChemicalsOutline', 'Thermo', 'IdealThermo', 
           'Stream', 'MultiStream', 'Heat', 'Power', 'ThermalCondition', 'ProcessSettings',
           'mixture', 'ThermoData', *reaction.__all__, *equilibrium.__all__,  *mixture.__all__,
           *network.__all__, 'preferences',
           'indexer', 'settings', 'functor', 'functors', 'chemicals', 'base', 
           'equilibrium', 'units_of_measure', 'exceptions', 'functional', 
           'reaction', 'constants', 'utils', 'separations', *phenomenode.__all__)

# Set number of digits displayed
import numpy as np
import pandas as pd
np.set_printoptions(suppress=False)
np.set_printoptions(precision=3) 
pd.options.display.float_format = '{:.3g}'.format
pd.set_option('display.max_rows', 50)
pd.set_option('display.max_columns', 10)
pd.set_option('max_colwidth', 35)
pd.set_option('display.width', 160)
del np, pd




# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .ideal import ideal
import thermosteam as tmo
from thermo.interaction_parameters import IPDB
from thermo import eos_mix
from fluids.constants import R_inv
import numpy as np

__all__ = ('FugacityCoefficients', 
           'IdealFugacityCoefficients')


class FugacityCoefficients:
    """
    Abstract class for the estimation of fugacity coefficients. Non-abstract subclasses should implement the following methods:
        
    __init__(self, chemicals: Iterable[:class:`~thermosteam.Chemical`]):
        Should use pure component data from chemicals to setup future calculations of fugacity coefficients.
    
    __call__(self, y: 1d array, T: float, P:float):
        Should accept an array of vapor molar compositions `y`, temperature `T` (in Kelvin), and pressure `P` (in Pascal), and return an array of fugacity coefficients. Note that the molar compositions must be in the same order as the chemicals defined when creating the FugacityCoefficients object.
        
    """
    __slots__ = ()
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"<{type(self).__name__}([{chemicals}])>"


@ideal
class IdealFugacityCoefficients(FugacityCoefficients):
    """
    Create an IdealFugacityCoefficients object that estimates all fugacity coefficients to be 1 when called with composition, temperature (K), and pressure (Pa).
    
    Parameters
    ----------
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals')
    
    def __init__(self, chemicals):
        self.chemicals = chemicals
    
    @property
    def chemicals(self):
        """tuple[Chemical] All chemicals involved in the calculation of fugacity coefficients."""
        return self._chemicals
    @chemicals.setter
    def chemicals(self, chemicals):
        self._chemicals = tuple(chemicals)

    def __call__(self, y, T, P):
        return 1.


class GCEOSFugacityCoefficients(FugacityCoefficients):
    """
    Abstract class for estimating fugacity coefficients using a generic cubic equation of state 
    when called with composition, temperature (K), and pressure (Pa).
    
    Parameters
    ----------
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals', '_eos')
    EOS = None # type[GCEOSMIX] Subclasses must implement this attribute.
    cache = None # [dict] Subclasses must implement this attribute.
    chemsep_db = None # Optional[str] Name of chemsep data base for interaction parameters.
    
    def __new__(cls, chemicals):
        chemicals = tuple(chemicals)
        cache = cls.cache
        if chemicals in cache:
            return cache[chemicals]
        else:
            self = super().__new__(cls)
            self._chemicals = chemicals
            cache[chemicals] = self
            return self
    
    @classmethod
    def subclass(cls, EOS, name=None):
        if name is None: name = EOS.__name__.replace('MIX', '') + 'FugacityCoefficients'
        return type(name, (cls,), dict(EOS=EOS, cache={}))
    
    @property
    def chemicals(self):
        """tuple[Chemical] All chemicals involved in the calculation of fugacity coefficients."""
        return self._chemicals
    
    def eos(self, zs, T, P):
        if zs.__class__ is np.ndarray: zs = [float(i) for i in zs]
        try:
            self._eos = eos = self._eos.to_TP_zs_fast(T=T, P=P, zs=zs, only_g=True, full_alphas=False)
        except:
            data = tmo.ChemicalData(self.chemicals)
            if self.chemsep_db is None:
                kijs = None
            else:
                try:
                    kijs = IPDB.get_ip_asymmetric_matrix(self.chemsep_db, data.CASs, 'kij')
                except:
                    kijs = None
            self._eos = eos = self.EOS(
                zs=zs, T=T, P=P, Tcs=data.Tcs, Pcs=data.Pcs, omegas=data.omegas, kijs=kijs,
                only_g=True
            )
        return eos
    
    def __call__(self, x, T, P):
        eos = self.eos(x, T, P)
        try:
            log_phis = np.array(eos.dlnphi_dns(eos.Z_g)) + eos.G_dep_g * R_inv / T
            return np.exp(log_phis)
        except:
            return 1.
    f = __call__
    args = ()

dct = globals()    
clsnames = []
fugacity_coefficient_classes = []
for name in ('PRMIX', 'SRKMIX', 'PR78MIX', 'VDWMIX', 'PRSVMIX',
             'PRSV2MIX', 'TWUPRMIX', 'TWUSRKMIX', 'APISRKMIX', 'IGMIX', 'RKMIX',
             'PRMIXTranslatedConsistent', 'PRMIXTranslatedPPJP', 'PRMIXTranslated',
             'SRKMIXTranslatedConsistent', 'PSRK', 'MSRKMIXTranslated',
             'SRKMIXTranslated'):
    cls = GCEOSFugacityCoefficients.subclass(getattr(eos_mix, name))
    clsname = cls.__name__
    clsnames.append(clsname)
    fugacity_coefficient_classes.append(cls)
    dct[clsname] = cls

dct['PRFugacityCoefficients'].chemsep_db = 'ChemSep PR'
__all__ = (*__all__, *clsnames, 'fugacity_coefficient_classes')
del dct, clsnames



# -*- coding: utf-8 -*-
"""
"""
import thermosteam as tmo
from chemicals.identifiers import pubchem_db

biorefinery_chemicals = {}

def search_biorefinery_chemicals(ID, error=None, **kwargs):
    if ID in biorefinery_chemicals: 
        f = biorefinery_chemicals[ID]
    else:
        key = ID.replace(' ', '').replace('-', '').replace('_', '').lower()
        if key in biorefinery_chemicals: 
            f = biorefinery_chemicals[key]
        elif error:
            raise error from None
        else:
            chemical = tmo.Chemical(ID, default=True, db='ChEDL', **kwargs)
            if chemical.Tm > 1000: chemical.at_state('s')
            return chemical
    return f(ID, default=True, **kwargs)

# Lignocellulosic heat capacities:
# Assume heat capacity of lignin, cellulose, and hemicellulose
# and all components at 350 K are about the same.
# https://link.springer.com/article/10.1007/s10853-013-7815-6
# https://www.sciencedirect.com/science/article/pii/0032386182901252
Cp_cellulosic = 1.364

# Assume density is similar for most solutes and solids, in agreement with Superpro models for sugarcane.
rho_solute = 1540 # kg/m3
rho_solid = 700 # kg/m3 - pelleted biomass

# Heats of formation for cellulosic components are from Humbird 2011 report: https://www.nrel.gov/docs/fy11osti/47764.pdf
# They are originally found in calories, so we need to convert them to joule.
cal2joule = 4.184

def _register(f, aliases):
    keys = [*aliases, f.__name__]
    keys.extend([i.lower() for i in keys])
    try:
        metadata = pubchem_db.search(f.__name__)
    except:
        pass    
    else:
        iupac_name = metadata.iupac_name, 
        if not iupac_name: iupac_name = ()
        elif isinstance(iupac_name, str): iupac_name = (iupac_name,)
        keys = set([*keys, metadata.common_name, metadata.formula])
    keys = set([i for i in keys if i not in biorefinery_chemicals])
    for i in keys: biorefinery_chemicals[i] = f
    return f

def register(*aliases):
    if len(aliases) == 1 and not isinstance(aliases[0], str):
        return _register(aliases[0], ())
    else:
        return lambda f: _register(f, aliases)

# %% Non-interacting solids

@register
def Ash(ID, **kwargs):
    return tmo.Chemical(ID, MW=1., phase='s', db=None,
                        Cp=0.37656, **kwargs)

@register
def Flocculant(ID, **kwargs):
    return tmo.Chemical(ID, MW=1., phase='s', db=None,
                        Cp=Cp_cellulosic, **kwargs)

@register
def Solids(ID, **kwargs):
    return tmo.Chemical(ID, MW=1., phase='s', db=None,
                        Cp=1.100, **kwargs)

# %% Microbes and cellular components

@register('DryYeast')
def Yeast(ID, **kwargs):
    return tmo.Chemical(ID, formula='CH1.61O0.56', rho=1112.6,
                        phase='s', db=None, 
                        Hf=(-7055.556554690305, 'J/g'),
                        aliases=['DryYeast', 'Yeast'], **kwargs)

@register('NYeast')
def NitrogenYeast(ID, **kwargs):
    return tmo.Chemical(ID, formula='CH1.61O0.56N0.16', phase='s',
                        rho=1112.6, db=None, 
                        Hf=(-7055.556554690305, 'J/g'),
                        aliases=['DryYeast', 'Yeast'], **kwargs)

@register('WWTsludge')
def Sluge(ID, **kwargs):
    return tmo.Chemical(ID, formula='CH1.64O0.39N0.23S0.0035', 
                        rho=1112.6, phase='s', 
                        Hf=-23200.01*cal2joule, db=None, **kwargs)

@register('Enzyme', 'Cellulase', 'Amylase')
def Protein(ID, **kwargs):
    return tmo.Chemical(ID, formula='CH1.57O0.31N0.29S0.007', 
                        rho=1112.6, phase='s', 
                        Hf=-17618*cal2joule, db=None, **kwargs)

# %% Cellulosic biomass components

@register('Mannose', 'Galactose')
def Glucose(ID, **kwargs):
    return tmo.Chemical(ID, search_ID='Glucose', phase='l', 
                        db='ChEDL', N_solutes=1, **kwargs)

@register
def Sucrose(ID, **kwargs):
    return tmo.Chemical(ID, search_ID='Sucrose', phase='l', 
                        db='ChEDL', N_solutes=2, **kwargs)

@register('Arabinose')
def Xylose(ID, **kwargs):
    return tmo.Chemical(ID, search_ID='Xylose', phase='l',
                        db='ChEDL', **kwargs)

@register('Cellulose')
def Glucan(ID, **kwargs):
    return tmo.Chemical(ID, formula='C6H10O5', phase='s', db=None,
                        Hf=-233200*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def Starch(ID, **kwargs):
    return tmo.Chemical(
        ID, formula='C6H10O5', phase='s', db=None,
        Hf=-233200*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
        **kwargs
    )

@register
def Fiber(ID, **kwargs):
    return tmo.Chemical(
        ID, formula='C6H10O5', phase='s', db=None,
        Hf=-233200*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
        **kwargs
    )

@register
def Hemicellulose(ID, **kwargs):
    return tmo.Chemical(ID, formula="C5H8O4", # Model formula as xylose monomer minus water
                        Hf=-761906.4, phase='s', db=None, 
                        Cp=Cp_cellulosic, **kwargs)

@register('Arabinan')
def Xylan(ID, **kwargs):
    return tmo.Chemical(ID, formula='C5H8O4', phase='s', db=None,
                        Hf=-182100*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def Lignin(ID, **kwargs):
    return tmo.Chemical(ID, formula='C8H8O3', phase='s', db=None,
                        Hf=-108248*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def SolubleLignin(ID, **kwargs):
    return tmo.Chemical(ID, formula='C8H8O3', phase='l', db=None,
                        Hf=-108248*cal2joule, rho=rho_solute, Cp=Cp_cellulosic,
                        **kwargs)

@register('GalactoseOligomer', 'MannoseOligomer')
def GlucoseOligomer(ID, **kwargs):
    return tmo.Chemical(ID, formula='C6H10O5', phase='l', db=None, 
                        Hf=-233200*cal2joule, rho=rho_solute, Cp=Cp_cellulosic, 
                        **kwargs)

@register('ArabinoseOligomer')
def XyloseOligomer(ID, **kwargs):
    return tmo.Chemical(ID, formula='C5H8O4', phase='l', db=None, 
                        Hf=-182100*cal2joule, rho=rho_solute, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def Xylitol(ID, **kwargs):
    return tmo.Chemical(ID, formula='C5H12O5', phase='l', db=None,
                        Hf=-243145*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def Cellobiose(ID, **kwargs):
    return tmo.Chemical(ID, formula='C12H22O11', phase='l', db=None,
                        Hf=-480900*cal2joule, rho=rho_solid, Cp=Cp_cellulosic, 
                        **kwargs)

@register
def H3PO4(ID, **kwargs):
    return tmo.Chemical(ID, search_ID='H3PO4', phase='s', db='ChEDL', **kwargs)

@register
def P4O10(ID, **kwargs):
    return tmo.Chemical(ID, search_ID='P4O10', phase='s', db='ChEDL', **kwargs)
 


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import chemicals_user
from . import thermo_user
from . import read_only
from . import registered
from . import forward
from . import define_units_of_measure
from . import chemical_cache
from . import jitdata
from . import jit_signature

__all__ = (*chemicals_user.__all__,
           *thermo_user.__all__,
           *read_only.__all__,
           *registered.__all__,
           *forward.__all__,
           *define_units_of_measure.__all__,
           *chemical_cache.__all__,
           *jitdata.__all__,
           *jit_signature.__all__)

from .registered import *
from .chemicals_user import *
from .thermo_user import *
from .read_only import *
from .forward import *
from .define_units_of_measure import *
from .chemical_cache import *
from .jitdata import *
from .jit_signature import *



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# A significant portion of this module originates from:
# Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
# 
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/thermo/blob/master/LICENSE.txt for details.
"""
"""
import numpy as np
import flexsolve as flx
from numba import njit

__all__ = ('phase_fraction', 
           'solve_phase_fraction_Rashford_Rice',
           'phase_composition',
           'compute_phase_fraction_2N')

@njit(cache=True)
def as_valid_fraction(x):
    """Ensure that x is between 0 and 1."""
    if x < 0.:
        x = 0.
    elif x > 1.:
        x = 1.
    return x

# @njit(cache=True)
def phase_fraction(zs, Ks, guess=None, za=0., zb=0.):
    """Return phase fraction for binary phase equilibrium."""
    N = zs.size
    if za or zb or N > 2:
        phase_fraction = solve_phase_fraction_Rashford_Rice(zs, Ks, guess, za, zb)
    else:
        if Ks.max() <= (1.0 + 1e-9): return 1
        if Ks.min() >= (1.0 - 1e-9): return 0
        if N == 2:
            phase_fraction = compute_phase_fraction_2N(zs, Ks)
        else:
            raise ValueError('number of chemicals in equilibrium must be 2 or more '
                             'to find phase fraction')
    return as_valid_fraction(phase_fraction)

@njit(cache=True)
def phase_composition(zs, Ks, phi):
    return zs * Ks / (phi * Ks + (1. - phi))

# @njit(cache=True)
def solve_phase_fraction_Rashford_Rice(zs, Ks, guess, za=0, zb=0):
    """
    Return phase fraction for N-component binary equilibrium by
    numerically solving the Rashford Rice equation.
    
    """
    if Ks.max() <= (1.0 + 1e-9) and not za: return 0.
    if Ks.min() >= (1.0 - 1e-9) and not zb: return 1.
    K_minus_1 = Ks - 1.
    args = (- zs * K_minus_1, K_minus_1, za, zb)
    f = phase_fraction_objective_function
    x0 = 1e-16 if za else 0.
    x1 = (1 - 1e-16) if zb else 1.
    y0 = f(x0, *args) 
    y1 = f(x1, *args)
    if y0 > y1 > 0.: return 1
    if y1 > y0 > 0.: return 0.
    if y0 < y1 < 0.: return 1.
    if y1 < y0 < 0.: return 0.
    x0, x1, y0, y1 = flx.find_bracket(f, x0, x1, y0, y1, args, tol=5e-8)
    if abs(x1 - x0) < 1e-6: return (x0 + x1) / 2.
    return flx.IQ_interpolation(f, x0, x1, y0, y1,
                                guess, 1e-16, 1e-16,
                                args, checkiter=False)

@njit(cache=True)
def phase_fraction_objective_function(phi, negative_zs_K_minus_1, K_minus_1, za, zb):
    """Phase fraction objective function."""
    denominator = 1. + phi * K_minus_1
    a = za/phi if za > 0. else 0.
    b = zb/(1. - phi) if zb > 0. else 0.
    phi = (negative_zs_K_minus_1 / denominator).sum()
    return phi - a + b

@njit(cache=True)
def compute_phase_fraction_2N(zs, Ks):
    """Return phase fraction for 2-component binary equilibrium."""
    z1, z2 = zs
    K1, K2 = Ks
    K1z1 = K1*z1
    K1z2 = K1*z2
    K2z1 = K2*z1
    K2z2 = K2*z2
    K1K2 = K1*K2
    K1K2z1 = K1K2*z1
    K1K2z2 = K1K2*z2
    z1_z2 = z1 + z2
    K1z1_K2z2 = K1z1 + K2z2
    return (-K1z1_K2z2 + z1_z2)/(K1K2z1 + K1K2z2 - K1z2 - K1z1_K2z2 - K2z1 + z1_z2)

# # @njit(cache=True)
# def solve_phase_fraction_iteration(zs, Ks, guess=0.5, za=0., zb=0.):
#     """
#     Return phase fraction for N-component binary phase equilibrium by 
#     accelerated fixed-point iteration. 
    
#     Notes
#     -----
#     This iterative method was developed by Yoel Cortes-Pena. It can handle 
#     chemicals which do not partition. za and zb are the fraction of 
#     non-partitioning chemicals in phases a and b, respectively. 
    
#     Examples
#     --------
#     Find the vapor fraction of a mixture of CO2, air, and water. Assume that 
#     air is always a gas, water is always a liquid, but CO2 can partition in both.
        
#     >>> import numpy as np
#     >>> from thermosteam.equilibrium import (
#     ...     solve_phase_fraction_iteration,
#     ... )
#     >>> F_air = 1
#     >>> F_water = 1
#     >>> F_CO2 = 1
#     >>> F_total = F_air + F_water + F_CO2
#     >>> z_air = F_air / F_total
#     >>> z_water = F_water / F_total
#     >>> zs = np.array([0.333]) # CO2
#     >>> Ks = np.array([0.999]) # CO2
#     >>> phi = solve_phase_fraction_iteration(
#     ...     zs, Ks, za=z_air, zb=z_water
#     ... )
#     >>> phi
#     0.4998
    
#     """
#     if Ks.max() < 1.0 and not za: return 0.
#     if Ks.min() > 1.0 and not zb: return 1.
#     args = (zs, Ks, za, zb)
#     x0 = 0.
#     x1 = 1.
#     f = compute_phase_fraction_iter
#     y0 = -np.inf if za else f(x0, *args) 
#     y1 = np.inf if zb else f(x1, *args)
#     if y0 > y1 > 0.: return 1
#     if y1 > y0 > 0.: return 0.
#     if y0 < y1 < 0.: return 1.
#     if y1 < y0 < 0.: return 0.
#     if not 0. < guess < 1.: guess = 0.5
#     phi = np.array([guess, 1. - guess])
#     zc = np.ones([2, 1]); zc[:, 0] = [za, zb]
#     N = zs.shape[0]
#     Ks_2d = np.ones([2, N])
#     zs_2d = np.ones([1, N])
#     zs_2d[0, :] = zs
#     Ks_2d[0, :] = Ks
#     Ks_2d[1, :] = 1. / Ks
#     phi = flx.wegstein(f, phi, 1e-16, 
#                        args=(zs_2d, Ks_2d, zc), checkiter=False)
#     return phi[0] / phi.sum()

# @njit(cache=True)
# def compute_phase_fraction_iter(phi, zs, Ks, zc):
#     ys = phase_composition(zs, Ks, phi)
#     new_phi = np.zeros(phi.shape)
#     shape = Ks.shape
#     M = shape[0]
#     N = shape[1]
#     for i in range(M):
#         isum = 0.
#         for j in range(N): isum += ys[i, j]
#         new_phi[i] = isum * phi[i] + zc[i, 0]
#     return new_phi



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import flexsolve as flx
from numba import njit
from .. import functional as fn
from ..exceptions import InfeasibleRegion
from .._settings import settings
from .domain import vle_domain

__all__ = ('DewPoint',)

# %% Solvers

# @njit(cache=True)
def gamma_iter(gamma, x_gamma, T, P, f_gamma, gamma_args):
    try:
        x = x_gamma / gamma
    except FloatingPointError:
        gamma[gamma < 1e-32] = 1e-32
        x = x_gamma / gamma
    # Add back trace amounts for activity coefficients at infinite dilution
    mask = x < 1e-32
    x[mask] = 1e-32
    x = fn.normalize(x)
    return f_gamma(x, T, *gamma_args, P=P)

def solve_x(x_guess, x_gamma, T, P, f_gamma, gamma_args):
    mask = x_guess < 1e-32
    x_guess[mask] = 1e-32
    x_guess = fn.normalize(x_guess)
    gamma = f_gamma(x_guess, T, *gamma_args, P=P)
    args = (x_gamma, T, P, f_gamma, gamma_args)
    gamma = flx.wegstein(
        gamma_iter, gamma, 1e-12, args=args, checkiter=False,
        checkconvergence=False, convergenceiter=5, maxiter=DewPoint.maxiter
    )
    try:
        return x_gamma / gamma
    except:
        return x_gamma / gamma_iter(gamma, *args)

# %% Dew point values container

class DewPointValues:
    __slots__ = ('T', 'P', 'IDs', 'z', 'x')
    
    def __init__(self, T, P, IDs, z, x):
        self.T = T
        self.P = P
        self.IDs = IDs
        self.z = z
        self.x = x
        
    @property
    def y(self): return self.z
        
    def __repr__(self):
        return f"{type(self).__name__}(T={self.T:.2f}, P={self.P:.0f}, IDs={self.IDs}, z={self.z}, x={self.x})"


class ReactiveDewPointValues:
    __slots__ = ('T', 'P', 'IDs', 'z0', 'dz', 'y', 'x')
    
    def __init__(self, T, P, IDs, z0, dz, y, x):
        self.T = T
        self.P = P
        self.IDs = IDs
        self.z0 = z0
        self.dz = dz
        self.y = y
        self.x = x
        
    def __repr__(self):
        return f"{type(self).__name__}(T={self.T:.2f}, P={self.P:.0f}, IDs={self.IDs}, z0={self.z0}, dz={self.dz}, y={self.y}, x={self.x})"


# %% Dew point calculation

class DewPoint:
    """
    Create a DewPoint object that returns dew point values when called with a 
    composition and either a temperture (T) or pressure (P).
    
    Parameters
    ----------
    chemicals=None : Iterable[:class:`~thermosteam.Chemical`], optional
    
    thermo=None : :class:`~thermosteam.Thermo`, optional
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
    >>> tmo.settings.set_thermo(chemicals)
    >>> DP = tmo.equilibrium.DewPoint(chemicals)
    >>> # Solve for dew point at constant temperautre
    >>> molar_composition = (0.5, 0.5)
    >>> dp = DP(z=molar_composition, T=355)
    >>> dp
    DewPointValues(T=355.00, P=92008, IDs=('Water', 'Ethanol'), z=[0.5 0.5], x=[0.849 0.151])
    >>> # Note that the result is a DewPointValues object which contain all results as attibutes
    >>> (dp.T, round(dp.P), dp.IDs, dp.z, dp.x)
    (355, 92008, ('Water', 'Ethanol'), array([0.5, 0.5]), array([0.849, 0.151]))
    >>> # Solve for dew point at constant pressure
    >>> DP(z=molar_composition, P=2*101324)
    DewPointValues(T=376.25, P=202648, IDs=('Water', 'Ethanol'), z=[0.5 0.5], x=[0.83 0.17])

    """
    __slots__ = ('chemicals', 'phi', 'gamma', 'IDs', 
                 'pcf', 'Psats', 'Tmin', 'Tmax', 'Pmin', 'Pmax')
    _cached = {}
    maxiter = 50
    T_tol = 1e-9
    P_tol = 1e-3
    Tmin_factor = 0.1
    Tmax_factor = 10
    Pmin_factor = 0.1 
    Pmax_factor = 10
    
    def __new__(cls, chemicals=None, thermo=None):
        thermo = settings.get_default_thermo(thermo)
        if chemicals is None:
            chemicals = thermo.chemicals.tuple
        else:
            chemicals = tuple(chemicals)
        key = (chemicals, thermo.Gamma, thermo.Phi, thermo.PCF)
        cached = cls._cached
        if key in cached:
            return cached[key]
        else:
            self = super().__new__(cls)
            self.IDs = tuple([i.ID for i in chemicals])
            self.gamma = thermo.Gamma(chemicals)
            self.phi = thermo.Phi(chemicals)
            self.pcf = thermo.PCF(chemicals)
            self.Psats = Psats = [i.Psat for i in chemicals]
            Tmin, Tmax = vle_domain(chemicals)
            self.Tmin = Tmin
            self.Tmax = Tmax
            self.Pmin = min([i(Tmin) for i in Psats])
            self.Pmax = max([i(Tmax) for i in Psats])
            self.chemicals = chemicals
            cached[key] = self
            return self
    
    def _solve_x(self, x_gamma, T, P, x):
        gamma = self.gamma
        return solve_x(x, x_gamma, T, P, gamma.f, gamma.args)
    
    def _T_error(self, T, P, z_norm, zP, x):
        if T <= 0: raise InfeasibleRegion('negative temperature')
        Psats = np.array([i(T) for i in self.Psats])
        Psats[Psats < 1e-16] = 1e-16 # Prevent floating point error
        phi = self.phi(z_norm, T, P)
        pcf = self.pcf(T, P, Psats)
        x_gamma = phi * zP / Psats / pcf
        x[:] = self._solve_x(x_gamma, T, P, x)
        return 1 - x.sum()
    
    def _T_error_reactive(self, T, P, z, dz, y, x, gas_conversion):
        if T <= 0: raise InfeasibleRegion('negative temperature')
        dz[:] = gas_conversion(z, T, P, 'g')
        y[:] = z + dz
        y /= y.sum()
        Psats = np.array([i(T) for i in self.Psats])
        Psats[Psats < 1e-16] = 1e-16 # Prevent floating point error
        phi = self.phi(y, T, P)
        pcf = self.pcf(T, P, Psats)
        x_gamma = phi * y * P / Psats / pcf
        x[:] = self._solve_x(x_gamma, T, P, x)
        return 1 - x.sum()
    
    def _T_error_ideal(self, T, zP, x):
        Psats = np.array([i(T) for i in self.Psats])
        Psats[Psats < 1e-16] = 1e-16 # Prevent floating point error
        x[:] = zP / Psats
        return 1 - x.sum()
    
    def _P_error(self, P, T, z_norm, z_over_Psats, Psats, x):
        if P <= 0: raise InfeasibleRegion('negative pressure')
        x_gamma = z_over_Psats * P * self.phi(z_norm, T, P) / self.pcf(T, P, Psats)
        x[:] = self._solve_x(x_gamma, T, P, x)
        return 1 - x.sum()
    
    def _P_error_reactive(self, P, T, Psats, z, dz, y, x, gas_conversion):
        if P <= 0: raise InfeasibleRegion('negative pressure')
        dz[:] = gas_conversion(z, T, P, 'g')
        y[:] = z + dz
        y /= y.sum()
        x_gamma = y / Psats * P * self.phi(y, T, P) / self.pcf(T, P, Psats)
        x[:] = self._solve_x(x_gamma, T, P, x)
        return 1 - x.sum()
    
    def _Tx_ideal(self, zP):
        f = self._T_error_ideal
        Tmin = self.Tmin + 10.
        Tmax = self.Tmax - 10.
        x = zP.copy() 
        args = (zP, x)
        fmin = f(Tmin, *args)
        if fmin > 0.: return Tmin, x
        fmax = f(Tmax, *args)
        if fmax < 0.: return Tmax, x
        T = flx.IQ_interpolation(f, Tmin, Tmax, fmin, fmax, 
                                 None, self.T_tol, 5e-12, args,
                                 checkiter=False, checkbounds=False,
                                 maxiter=self.maxiter)
        return T, x
    
    def _Px_ideal(self, z_over_Psats):
        P = 1. / z_over_Psats.sum()
        x = z_over_Psats * P
        return P, x
    
    def __call__(self, z, *, T=None, P=None, gas_conversion=None, guess=None):
        z = np.asarray(z, float)
        if T:
            if P: raise ValueError("may specify either T or P, not both")
            P, *args = self.solve_Px(z, T, gas_conversion, guess)
        elif P:
            T, *args = self.solve_Tx(z, P, gas_conversion, guess)
        else:
            raise ValueError("must specify either T or P")
        if gas_conversion:
            return ReactiveDewPointValues(T, P, self.IDs, z, *args)
        else:
            return DewPointValues(T, P, self.IDs, z, *args)
    
    def solve_Tx(self, z, P, gas_conversion=None, guess=None):
        """
        Dew point given composition and pressure.

        Parameters
        ----------
        z : ndarray
            Molar composition.
        P : float
            Pressure [Pa].

        Returns
        -------
        T : float
            Dew point temperature [K].
        x : numpy.ndarray
            Liquid phase molar composition.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> import numpy as np
        >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> DP = tmo.equilibrium.DewPoint(chemicals)
        >>> tmo.docround(DP.solve_Tx(z=np.array([0.5, 0.5]), P=101325))
        (357.4419, array([0.847, 0.153]))
        
        """
        positives = z > 0.
        N = positives.sum()
        if N == 0:
            raise ValueError('no positive components present')
        if N == 1:
            chemical = self.chemicals[fn.first_true_index(positives)]
            T = chemical.Tsat(P, check_validity=False) if P <= chemical.Pc else chemical.Tc
            x = z.copy()
            return T, fn.normalize(x)
        elif gas_conversion is None:
            f = self._T_error
            z_norm = z/z.sum()
            zP = z * P
            if guess is None:
                T_guess0, x = self._Tx_ideal(zP) 
                T_guess1 = T_guess0 + 1e-3
            else:
                T_guess0, x, T_guess1 = guess
            args = (P, z_norm, zP, x)
            try:
                T = flx.aitken_secant(f, T_guess0, T_guess1,
                                      self.T_tol, 5e-12, args,
                                      maxiter=self.maxiter,
                                      checkiter=False)
            except RuntimeError:
                Tmin = max(self.Tmin_factor * T_guess0, self.Tmin)
                Tmax = min(self.Tmax_factor * T_guess0, self.Tmax)
                T = flx.IQ_interpolation(f, Tmin, Tmax,
                                         f(Tmin, *args), f(Tmax, *args),
                                         T_guess0, self.T_tol, 5e-12, args,
                                         checkiter=False, checkbounds=False,
                                         maxiter=self.maxiter)
            return T, fn.normalize(x)
        else:
            f = self._T_error_reactive
            z_norm = z / z.sum()
            x = z_norm.copy()
            dz = z_norm.copy()
            zP = z * P
            T_guess, y = self._Tx_ideal(zP)
            args = (P, z_norm, dz, y, x, gas_conversion)
            try:
                T = flx.aitken_secant(f, T_guess, T_guess + 1e-3,
                                      self.T_tol, 5e-12, args,
                                      checkiter=False)
            except RuntimeError:
                Tmin = max(self.Tmin_factor * T_guess, self.Tmin)
                Tmax = min(self.Tmax_factor * T_guess, self.Tmax)
                T = flx.IQ_interpolation(f, Tmin, Tmax,
                                         f(Tmin, *args), f(Tmax, *args),
                                         T_guess, self.T_tol, 5e-12, args, 
                                         checkiter=False, checkbounds=False)
            return T, dz, fn.normalize(y), x
    
    def solve_Px(self, z, T, gas_conversion=None, guess=None):
        """
        Dew point given composition and temperature.

        Parameters
        ----------
        z : ndarray
            Molar composition.
        T : float
            Temperature (K).
        
        Returns
        -------
        P : float
            Dew point pressure (Pa).
        x : ndarray
            Liquid phase molar composition.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> import numpy as np
        >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> DP = tmo.equilibrium.DewPoint(chemicals)
        >>> tmo.docround(DP.solve_Px(z=np.array([0.5, 0.5]), T=352.28))
        (82480.7311, array([0.851, 0.149]))
 
        """
        positives = z > 0.
        N = positives.sum()
        if N == 0:
            raise ValueError('no positive components present')
        if N == 1:
            chemical = self.chemicals[fn.first_true_index(z)]
            P = chemical.Psat(T) if T <= chemical.Tc else chemical.Pc
            x = z.copy()
            return P, fn.normalize(x)
        elif gas_conversion is None:
            z_norm = z / z.sum()
            Psats = np.array([i(T) for i in self.Psats], dtype=float)
            z_over_Psats = z / Psats
            if guess is None:
                P_guess0, x = self._Px_ideal(z_over_Psats)
                P_guess1 = P_guess0 - 10
            else:
                P_guess0, x, P_guess1 = guess
            args = (T, z_norm, z_over_Psats, Psats, x)
            f = self._P_error
            try:
                P = flx.aitken_secant(f, P_guess0, P_guess1, self.P_tol, 5e-12, args,
                                      checkiter=False, maxiter=self.maxiter)
            except RuntimeError:
                Pmin = max(self.Pmin_factor * P_guess0, self.Pmin)
                Pmax = min(self.Pmax_factor * P_guess0, self.Pmax)
                P = flx.IQ_interpolation(f, Pmin, Pmax, 
                                         f(Pmin, *args), f(Pmax, *args),
                                         P_guess0, self.P_tol, 5e-12, args,
                                         checkiter=False, checkbounds=False,
                                         maxiter=self.maxiter)
            return P, fn.normalize(x)
        else:
            f = self._P_error_reactive
            z_norm = z / z.sum()
            y = z_norm.copy()
            dz = z_norm.copy()
            Psats = np.array([i(T) for i in self.Psats], dtype=float)
            z_over_Psats = z / Psats
            P_guess, x = self._Px_ideal(z_over_Psats)
            args = (T, Psats, z_norm, dz, y, x, gas_conversion)
            try:
                P = flx.aitken_secant(f, P_guess, P_guess-1, self.P_tol, 1e-9,
                                      args, checkiter=False)
            except RuntimeError:
                Pmin = max(self.Pmin_factor * P_guess, self.Pmin)
                Pmax = min(self.Pmax_factor * P_guess, self.Pmax)
                P = flx.IQ_interpolation(f, Pmin, Pmax,
                                         f(Pmin, *args), f(Pmax, *args),
                                         P_guess, self.P_tol, 5e-12, args,
                                         checkiter=False, checkbounds=False)
            return P, dz, fn.normalize(y), x
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"{type(self).__name__}([{chemicals}])"
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
from .._phase import check_phase

__all__ = ('PartitionCoefficients',)


class PartitionCoefficients:
    """
    Create a PartitionCoefficients capable of computing partition coefficients 
    when called with a composition vectors (1d arrays) for the two phases, the temperature,
    and the pressure.
    
    Parameters
    ----------
    phases : str, optional
        Phases in equilibrium.
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
        Chemicals to compute fugacities.
    thermo : :class:`~thermosteam.Thermo`, optional
        Thermodynamic property package.
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> import numpy as np
    >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'])
    >>> tmo.settings.set_thermo(chemicals)
    >>> # Create a PartitionCoefficients object
    >>> K_gl = tmo.equilibrium.PartitionCoefficients('gl', chemicals)
    >>> K_gl
    PartitionCoefficients('gl', [Water, Ethanol])
    >>> # Compute partition coefficients
    >>> liquid_molar_composition = np.array([0.72, 0.28])
    >>> vapor_molar_composition = np.array([0.28, 0.72])
    >>> k_gl = K_gl(y=vapor_molar_composition, x=liquid_molar_composition, T=355, P=101325)
    >>> k_gl
    array([0.594, 2.035])
    
    """
    __slots__ = ('phases', 'fugacities_by_phase')
    
    def __init__(self, phases, chemicals, thermo=None):
        if len(phases) != 2: raise ValueError('phases argument must consist of two phases')
        for i in phases: check_phase(i)
        self.phases = phases
        self.fugacities_by_phase = tmo.equilibrium.fugacities_by_phase(chemicals, thermo)
    
    def __call__(self, y, x, T, P):
        top_phase, bottom_phase = self.phases
        return (
            self.fugacities_by_phase[bottom_phase].unweighted(x, T, P)
            / self.fugacities_by_phase[top_phase].unweighted(y, T, P)
        )
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.fugacities_by_phase[self.phases[0]].chemicals])
        return f"{type(self).__name__}({self.phases!r}, [{chemicals}])"




# -*- coding: utf-8 -*-
"""
"""
from thermo.interaction_parameters import IPDB
import thermosteam as tmo
from thermo.bulk import default_settings
import thermo as tm

__all__ = (
    # 'FlashPackage',
)

# class FlashPackage:
#     """
#     Create a FlashPackage object that predefines flash algorithms
#     for easier creation of thermo Flash and Phase objects.
    
#     Parameters
#     ----------
#     G : :obj:`Phase subclass <thermo.Phase>`, optional
#         Class to create gas phase object. Defaults to `CEOSGas <thermo.CEOSGas>`.
#     L : :obj:`Phase subclass <thermo.Phase>`, optional
#         Class to create liquid phase object. Defaults to :obj:`GibbsExcessLiquid <thermo.GibbsExcessLiquid>`.
#     S : :obj:`Phase subclass <thermo.Phase>`, optional
#         Class to create solid phase object. Defaults to :obj:`GibbsExcessSolid <thermo.GibbsExcessSolid>`.
#     GE : :obj:`GibbsExcessModel subclass <thermo.Phase>`, optional
#         Class to create GibbsExcessModel object. Defaults to :obj:`UNIFAC <thermo.UNIFAC>`.
#     Gkw : dict, optional
#         Key word arguments to initialize `G`.
#     Lkw : dict, optional
#         Key word arguments to initialize `L`.
#     Skw : dict, optional
#         Key word arguments to initialize `S`.
#     GEkw : dict, optional
#         Key word arguments to initialize `GE`.
#     settings : :obj:`BulkSettings <thermo.bulk.BulkSettings>`, optional
#         Object containing settings for calculating bulk and transport
#         properties, [-]
#     chemicals : :obj:`Chemicals <thermosteam.chemicals.Chemicals>`, optional
#         Chemicals available to flash. Defaults to thermosteam.settings.get_chemicals()
#     N_liquid : int, optional
#         Default number of liquid phases. Defaults to 1.
#     N_solid : int, optional
#         Default number of solid phases. Defaults to 0.
        
#     Examples
#     --------
#     >>> import thermo as tm
#     >>> import thermosteam as tmo
#     >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Hexane'])
#     >>> # VLLE using activity coefficients for the liquid phase 
#     >>> # and equations of state for the gas phase.
#     >>> flashpkg = tmo.equilibrium.FlashPackage(
#     ...     G=tm.CEOSGas, L=tm.GibbsExcessLiquid, S=tm.GibbsExcessSolid,
#     ...     GE=tm.UNIFAC, GEkw=dict(version=1), Gkw=dict(eos_class=tm.PRMIX),
#     ... )
#     >>> flasher = flashpkg.flasher(N_liquid=2)
#     >>> type(flasher).__name__
#     'FlashVLN'
#     >>> PT = flasher.flash(zs=[0.3, 0.2, 0.5], T=330, P=101325)
#     >>> tmo.docround([PT.VF, PT.betas, PT.liquid0.zs, PT.H()]) # doctest: +SKIP
#     [0.7514, [0.7514, 0.2486], [0.7606, 0.2347, 0.0046], -7650.0958]
#     >>> # VLE using activity coefficients for the liquid phase 
#     >>> # and equations of state for the gas phase.
#     >>> flasher = flashpkg.flasher(['Water', 'Ethanol'], N_liquid=1)
#     >>> type(flasher).__name__
#     'FlashVL'
#     >>> PT = flasher.flash(zs=[0.5, 0.5], T=353, P=101325)
#     >>> tmo.docround([PT.VF, PT.gas.zs, PT.H()]) # doctest: +SKIP
#     [0.2738, [0.3614, 0.6386], -27042.1818]
#     >>> # Single component equilibrium.
#     >>> flasher = flashpkg.flasher(['Ethanol']) 
#     >>> type(flasher).__name__
#     'FlashPureVLS'
#     >>> PT = flasher.flash(T=353, P=101325)
#     >>> tmo.docround([PT.VF, PT.gas.zs, PT.H()]) # doctest: +SKIP
#     [1.0, [1.0], 3619.7887]
#     >>> # VLLE using virial equation of state for the gas phase
#     >>> flashpkg.G, flashpkg.Gkw = tm.VirialGas, {}
#     >>> flasher = flashpkg.flasher(N_liquid=2)
#     >>> PT = flasher.flash(zs=[0.45, 0.05, 0.5], T=330, P=101325)
#     >>> tmo.docround([PT.VF, PT.betas, PT.liquid0.zs, PT.H(), PT.phase_count]) # doctest: +SKIP
#     [0.0, [0.4754, 0.5246], [0.9183, 0.0808, 0.0008], -33426.8214, 2]
#     >>> # VLLE using SRK equation of state for the gas phase
#     >>> flashpkg.G, flashpkg.Gkw = tm.CEOSGas, {'eos_class': tm.SRKMIX}
#     >>> flasher = flashpkg.flasher(N_liquid=2)
#     >>> PT = flasher.flash(zs=[0.45, 0.05, 0.5], T=330, P=101325)
#     >>> tmo.docround([PT.VF, PT.betas, PT.liquid0.zs, PT.H(), PT.phase_count]) # doctest: +SKIP
#     [0.0, [0.4754, 0.5246], [0.9183, 0.0808, 0.0008], -33426.8213, 2]
#     >>> # VLLE using ideal gas
#     >>> flashpkg.G, flashpkg.Gkw = tm.IdealGas, {}
#     >>> flasher = flashpkg.flasher(N_liquid=2)
#     >>> PT = flasher.flash(zs=[0.45, 0.05, 0.5], T=330, P=101325)
#     >>> tmo.docround([PT.VF, PT.betas, PT.liquid0.zs, PT.H()]) # doctest: +SKIP
#     [0.0, [0.4754, 0.5246], [0.9183, 0.0808, 0.0008], -33426.8213]
    
#     """
#     __slots__ = (
#         'chemicals', 'settings',
#         'G', 'Gkw',
#         'L', 'Lkw', 
#         'S', 'Skw',
#         'GE', 'GEkw',
#         'N_liquid', 'N_solid',
#     )
    
#     def __init__(self, G=tm.CEOSGas, L=tm.GibbsExcessLiquid, S=tm.GibbsExcessSolid,
#                  GE=tm.UNIFAC, Gkw=None, Lkw=None, Skw=None, GEkw=None, 
#                  settings=None, chemicals=None, N_liquid=None, N_solid=None):
#         self.chemicals = tmo.settings.get_chemicals() if chemicals is None else chemicals
#         self.G = G
#         self.L = L
#         self.S = S
#         self.GE = GE
#         self.Gkw = Gkw or {}
#         self.Lkw = Lkw or {}
#         self.Skw = Skw or {}
#         self.GEkw = GEkw or {}
#         self.settings = settings or default_settings
#         self.N_liquid = N_liquid or 2
#         self.N_solid = N_solid or 0
        
#     @property
#     def data(self):
#         return tmo.ChemicalData(self.chemicals)
    
#     def flasher(self, IDs=None, N_liquid=None, N_solid=None):
#         return self._flash_from_data(
#             self.data[IDs] if IDs else self.data,
#             N_liquid, N_solid
#         )

#     def _solid_from_data(self, data):
#         raise NotImplementedError("this method is not implemented yet")
        
#     def _liquid_from_data(self, data):
#         if self.L is tm.GibbsExcessLiquid and self.GE:
#             GE = self.GE.from_data(data, **self.GEkw)
#             return self.L.from_data(data, GibbsExcessModel=GE, **self.Lkw)
#         else:
#             return self.L.from_data(data, **self.Lkw)

#     def _gas_from_data(self, data):
#         return self.G.from_data(data, **self.Gkw)

#     def _flash_from_data(self, data, N_liquid, N_solid):
#         N = len(data.CASs)
#         if N_solid is None: N_solid = self.N_solid
#         if N_liquid is None: N_liquid = self.N_liquid
#         if N == 0:
#             raise ValueError(
#                 "IDs cannot be empty; at least one component ID must be given"
#             )
#         elif N == 1: # Pure component
#             return tm.FlashPureVLS(
#                 data, 
#                 data,
#                 self._gas_from_data(data),
#                 [self._liquid_from_data(data)
#                   for i in range(N_liquid)],
#                 [self._solid_from_data(data)
#                   for i in range(N_solid)],
#                 self.settings,
#             )
#         elif N_liquid == 1:
#             if N_solid:
#                 raise NotImplementedError(
#                     'multi-component flasher with solid phases '
#                     'not implemented (yet)'
#                 )
#             return tm.FlashVL(
#                 data, 
#                 data,
#                 self._gas_from_data(data),
#                 self._liquid_from_data(data),
#                 self.settings,
#             )
#         else:
#             if N_solid:
#                 raise NotImplementedError(
#                     'multi-component flasher with solid phases '
#                     'not implemented (yet)'
#                 )
#             return tm.FlashVLN(
#                 data, 
#                 data,
#                 [self._liquid_from_data(data)
#                   for i in range(N_liquid)],
#                 self._gas_from_data(data),
#                 [],
#                 self.settings,
#             )
 
def constructor(cls):
    return lambda f: setattr(cls, f.__name__, classmethod(f))
        
@constructor(tm.GibbsExcessLiquid)
def from_data(cls, data,
            GibbsExcessModel=None,
            use_Hvap_caloric=False,
            use_Poynting=False,
            use_phis_sat=False,
            use_Tait=False,
            use_eos_volume=False,
            Psat_extrapolation='AB',
            equilibrium_basis=None,
            caloric_basis=None):
    return cls(
        data.VaporPressures, 
        VolumeLiquids=data.VolumeLiquids,
        HeatCapacityGases=data.HeatCapacityGases,
        GibbsExcessModel=GibbsExcessModel,
        eos_pure_instances=[i.eos for i in data.VolumeGases],
        EnthalpyVaporizations=data.EnthalpyVaporizations,
        HeatCapacityLiquids=data.HeatCapacityLiquids,
        use_Hvap_caloric=use_Hvap_caloric,
        use_Poynting=use_Poynting,
        use_phis_sat=use_phis_sat,
        use_Tait=use_Tait,
        use_eos_volume=use_eos_volume,
        equilibrium_basis=equilibrium_basis,
        caloric_basis=caloric_basis
    )

@constructor(tm.CEOSGas)
def from_data(cls, data, eos_class=None):
    if eos_class is None: eos_class = tm.PRMIX
    eos_kwargs = dict(Tcs=data.Tcs,
                      Pcs=data.Pcs,
                      omegas=data.omegas)
    try:
        eos_kwargs['kijs'] = IPDB.get_ip_asymmetric_matrix('ChemSep PR', data.CASs, 'kij')
    except:
        pass        
    return cls(eos_class, eos_kwargs, data.HeatCapacityGases)

@constructor(tm.CEOSLiquid)
def from_data(cls, data, eos_class=None):
    if eos_class is None: eos_class = tm.PRMIX
    eos_kwargs = dict(Tcs=data.Tcs,
                      Pcs=data.Pcs,
                      omegas=data.omegas)
    try:
        eos_kwargs['kijs'] = IPDB.get_ip_asymmetric_matrix('ChemSep PR', data.CASs, 'kij')
    except:
        pass        
    return cls(eos_class, eos_kwargs, data.HeatCapacityGases)

tm.VirialGas.model_attributes = ('HeatCapacityGases', 'model') # TODO: Add attribute in Caleb's thermo
@constructor(tm.VirialGas)
def from_data(cls, data, model=None):
    if model is None:
        model = tm.VirialCSP(Tcs=data.Tcs, Pcs=data.Pcs, Vcs=data.Vcs, omegas=data.omegas)
    return cls(model, HeatCapacityGases=data.HeatCapacityGases)

@constructor(tm.IdealGas)
def from_data(cls, data):
    return cls(HeatCapacityGases=data.HeatCapacityGases)

@constructor(tm.UNIFAC)
def from_data(cls, data, version=1):
    if version == 0:
        chemgroups = data.UNIFAC_groups
    elif version == 1:
        chemgroups = data.UNIFAC_Dortmund_groups
    elif version == 2:
        chemgroups = data.PSRK_groups
    else:
        raise RuntimeError('chemgroups for version %d not yet implemented' %version)
    N = len(chemgroups)
    return cls.from_subgroups(298.15, N*[1./N], chemgroups, version=version)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from warnings import warn

__all__ = ('Registry', 'is_valid_ID', 'check_valid_ID')


def is_valid_ID(ID):
    return isinstance(ID, str)

def check_valid_ID(ID):
    if not isinstance(ID, str):
        raise RuntimeError(f"ID must be a string, not a '{type(ID).__name__}' object")
        
def check_valid_alias(alias):
    if not isinstance(alias, str):
        raise RuntimeError(f"alias must be a string, not a '{type(alias).__name__}' object")
    if not alias[0].isalpha():
        raise RuntimeError("alias must start with a letter")

class Registry: # pragma: no cover

    __slots__ = ('data', 'safe_to_replace', 'context_levels', 'registered_objects')

    AUTORENAME = False #: Whether to rename objects with conflicting IDs

    def untrack(self, objs):
        """
        Mark objects safe to replace, so no warnings are issued if any are 
        replaced.
        
        """
        self.safe_to_replace.update(objs)

    def track(self, objs):
        """
        Reregister objects if they are not already in the registry and mark 
        objects as unsafe to replace so that warnings can be issued if any are 
        replaced.
        
        Warning
        -------
        Overrides old objects with the same ID.
        
        """
        safe_to_replace = self.safe_to_replace
        data = self.data
        for obj in objs:
            safe_to_replace.discard(obj)
            data[obj._ID] = obj
        
    def __init__(self, objs=None):
        self.data = {i.ID: i for i in objs} if objs else {}
        self.safe_to_replace = set()
        self.context_levels = []
        self.registered_objects = set()

    def search(self, ID):
        """Return object given ID. If ID not in registry, return None."""
        return self.data.get(ID)
    
    def __dir__(self):
        return [*self.data, 'search', 'register', 'register_safely', 'discard', 
                'clear', 'mark_safe_to_replace', 'unmark_safe_to_replace']
    
    def __getattr__(self, ID):
        try:
            return self.data[ID]
        except KeyError:
            raise AttributeError(f'{repr(ID)} not in registry')
    __getitem__ = __getattr__
    
    def register_safely(self, ID, obj):
        """Register object safely, with checks and due warnings."""
        check_valid_ID(ID)
        data = self.data
        ID_old = self._open_registration(ID, obj)
        if ID in data:
            other = data[ID]
            if obj is not other and other not in self.safe_to_replace:
                if self.AUTORENAME:
                    try:
                        base, num = ID.rsplit('_', 1)
                        num = int(num)
                    except:
                        other.ID = ID + '_1'
                    else:
                        other.ID = base + '_' + str(num + 1) 
                    base, num = other.ID.rsplit('_', 1)
                    ID = base + '_' + str(int(num) + 1)
                elif ID_old:
                    warning = RuntimeWarning(
                        f"upon renaming, {obj!r} replaced {other!r} "
                         "in registry"
                    )
                    warn(warning, 4)
                else:
                    warning = RuntimeWarning(
                        f"{other!r} has been replaced in registry"
                    )
                    warn(warning, stacklevel=getattr(obj, '_stacklevel', 4) - 1)
        self._close_registration(ID, obj)
        
    def register(self, ID, obj):
        """Register object without warnings or checks."""
        self._open_registration(ID, obj)
        self._close_registration(ID, obj)
        
    def register_alias_safely(self, alias, obj, override=None):
        """Register object safely, with checks and due warnings."""
        check_valid_alias(alias)
        data = self.data
        if alias in data:
            other = data[alias]
            if obj is not other and other not in self.safe_to_replace:
                if override is None:
                    warning = RuntimeWarning(
                        f"alias {alias!r} of {other!r} has been replaced in registry with {obj!r}"
                    )
                    warn(warning, stacklevel=getattr(obj, '_stacklevel', 5) - 1)
                elif not override:
                    raise RuntimeError(
                        f"alias {alias!r} is already registered to {other!r}"
                    )
        self.register_alias(alias, obj)
        
    def register_alias(self, alias, obj, override=True):
        """Register object alias without warnings or checks."""
        if override:
            self.data[alias] = obj
        elif alias not in self.data:
            self.data[alias] = obj
        for i in self.context_levels: i.append(obj)
        
    def _open_registration(self, ID, obj):
        data = self.data
        ID_old = getattr(obj, '_ID', None)
        if ID_old and data.get(ID_old) is obj: del data[ID_old]
        return ID_old
        
    def _close_registration(self, ID, obj):
        if obj not in self.registered_objects:
            self.registered_objects.add(obj)
            for i in self.context_levels: 
                i.append(obj)
        self.data[ID] = obj
        obj._ID = ID
    
    def open_context_level(self):
        self.context_levels.append([])
        
    def close_context_level(self):
        return self.context_levels.pop()
    
    def get_context_level(self, level=None):
        if level is None:
            objs = list(self.data.values())
        else:
            context_levels = self.context_levels
            N_levels = len(context_levels)
            if (level >= 0 and level >= N_levels
                or level < 0 and -level > N_levels):
                objs = list(self.data.values())
            else:
                objs = context_levels[level]
        return objs
    
    def clear(self):
        """Clear data."""
        self.data.clear()
    
    def discard(self, obj):
        """Remove object from data."""
        data = self.data
        if hasattr(obj, '_ID'):
            ID = obj._ID
            if ID in data and data[ID] is obj: del data[ID]
        elif isinstance(obj, str):
            if obj in data: 
                ID = obj
                obj = data[obj]
                del data[ID]
        for dump in self.context_levels:
            if obj in dump: dump.remove(obj)
        self.registered_objects.discard(obj)
    
    def pop(self, obj):
        """Remove object from data and return object."""
        data = self.data
        if hasattr(obj, '_ID'):
            ID = obj._ID
            if ID in data and data[ID] is obj: del data[ID]
        elif isinstance(obj, str):
            if obj in data: 
                ID = obj
                obj = data[obj]
                del data[ID]
        for dump in self.context_levels:
            if obj in dump: dump.remove(obj)
        return obj
    
    def __contains__(self, obj):
        data = self.data
        if hasattr(obj, '_ID'):
            ID = obj._ID
            return ID in data and data[ID] is obj
        else:
            return obj in data
    
    def __iter__(self):
        return iter(self.data.values())
    
    def __repr__(self):
        return f"{type(self).__name__}([{', '.join([str(i) for i in self])}])"
    
    def show(self):
        if self.data:
            print(f'{type(self).__name__}:\n' + '\n'.join([repr(i) for i in self]))
        else:
            print(f'{type(self).__name__}: (Empty)')
    
    _ipython_display_ = show




# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the elements module from the chemicals library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from chemicals import elements, periodic_table
import numpy as np

elements.__all__.extend([
    'atoms_to_array', 'array_to_atoms',
])

#: Dict[str, dict[str, int]] Cache of atomic counts.
formula_to_atoms = {}

#: Dict[str, int] Symbol - index pairs for atomic arrays.
symbol_to_index = {e.symbol: e.number - 1 for e in periodic_table}

#: tuple[str] Symbols for atomic arrays.
symbols = tuple(symbol_to_index)

def atoms_to_array(atoms: dict) -> np.ndarray:
    index = symbol_to_index
    array = np.zeros(118)
    for symbol, value in atoms.items():
        array[index[symbol]] = value
    return array

def array_to_atoms(array: np.ndarray) -> dict:
    index, = np.where(array != 0.)
    return dict(zip([symbols[i] for i in index], array[index]))


def get_atoms(formula):
    if formula in formula_to_atoms:
        return formula_to_atoms[formula]
    else:
        formula_to_atoms[formula] = atoms = elements.simple_formula_parser(formula)
        if len(formula_to_atoms) > 50: del formula_to_atoms[next(iter(formula_to_atoms))]
    return atoms.copy() # Prevent cached atoms from being altered

elements.get_atoms = get_atoms 
elements.atoms_to_array = atoms_to_array
elements.array_to_atoms = array_to_atoms
elements.symbol_to_index = symbol_to_index
elements.symbols = symbols


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from math import pi

__all__ = ('R', 'k', 'g', 'pi', 'N_A', 'calorie', 'epsilon_0')


# Obtained from SciPy 0.19 (2014 CODATA)
# Included here so calculations are consistent across SciPy versions

#: Universal gas constant [J/mol/K]
R = 8.3144598

#: Boltzman constant [m^2 * kg * s^-2 * K^-1]
k = 1.38064852e-23

#: Avogadro's number
N_A = 6.022140857e+23

#: Conversion factor [J/calorie]
calorie = 4.184

#: Permittivity of a vacuum [C/V/m]
epsilon_0 = 8.854187817620389e-12

#: Acceleration due to gravity [m/s^2]
g = 9.80665

#: Planck's constant [m^2 * kg / s]
h = 6.62607004e-34

#: Speed of light in a vacuum [m /2]
c = 299792458


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

import matplotlib.pyplot as plt
from typing import Iterable
import matplotlib.ticker as ticker
import numpy as np

__all__ = (
    'set_font',
    'set_figure_size',
    'style_axis',
    'style_plot_limits',
    'fill_plot',
    'set_axes_labels',
    'set_axes_xlabels',
    'set_axes_ylabels',
)   

plt.rcParams.update({
    "figure.facecolor": (1, 1, 1, 0.95),
    "axes.facecolor": (1, 1, 1, 0.95),
    "legend.facecolor": (1, 1, 1, 0.95),
    "savefig.facecolor": (1, 1, 1, 0.95),
})

def set_font(size=8, family='sans-serif', font='Arial'):
    import matplotlib
    fontkwargs = {'size': size}
    matplotlib.rc('font', **fontkwargs)
    params = matplotlib.rcParams
    params['font.' + family] = font
    params['font.family'] = family
    
def set_figure_size(width=None, aspect_ratio=None, units=None): 
    # units default to inch
    # width defaults 6.614 inches
    # aspect ratio defaults to 0.65
    if aspect_ratio is None:
        aspect_ratio = 0.65
    if width is None or width == 'full':
        width = 6.6142
    elif width == 'half':
        width = 6.6142 / 2
    else:
        if units is not None:
            from thermosteam.units_of_measure import convert
            width = convert(width, units, 'inch')
    import matplotlib
    params = matplotlib.rcParams
    params['figure.figsize'] = (width, width * aspect_ratio)

def set_ticks(ax, ticks, which='x', ticklabels=(),
              labelrotation=0., ha=None, offset=False, 
              fs=None, ticklen=None):
    if ticklen is None: ticklen = 4
    if which == 'x':
        axis = ax.xaxis
        set_ticks = plt.xticks
        kwargs = dict(top=False)
    elif which == 'y':
        axis = ax.yaxis
        set_ticks = plt.yticks
        kwargs = dict(right=False)
    else:
        raise ValueError("which must be either 'x' or 'y'")
    if offset:
        ticks = np.array(ticks, float)
        ticks_offset = np.zeros(len(ticks) - 1)
        if len(ticks_offset) != 0:
            ticks_offset[:] = ticks[:-1] + 0.5 * np.diff(ticks)
        ax.tick_params(axis=which, which='major', length=ticklen,
                       direction="inout", **kwargs)
        ax.tick_params(axis=which, which='minor', length=0,
                       labelrotation=labelrotation, **kwargs)
        set_ticks(ticks_offset, (), fontsize=fs)
        if ticklabels:
            axis.set_minor_locator(ticker.FixedLocator(ticks))
            axis.set_minor_formatter(ticker.FixedFormatter(ticklabels))
        ticks = ticks_offset
    else:
        set_ticks(ticks, ticklabels, fontsize=fs)
        ax.tick_params(axis=which, direction="inout", length=ticklen, 
                       labelrotation=labelrotation, **kwargs)
    if ha is not None:
        for i in axis.get_majorticklabels():
            i.set_ha(ha)
    return ticks
    

def style_axis(ax=None, xticks=None, yticks=None, 
               xticklabels=None, yticklabels=None,
               top=True, right=True, trim_to_limits=False,
               xtick0=True, ytick0=True,
               xtickf=True, ytickf=True,
               offset_xticks=False,
               offset_yticks=False,
               xrot=None, xha=None,
               yrot=None, yha=None,
               axtwinx=None, axtwiny=None,
               fs=None, ticklen=None): # pragma: no cover
    if ticklen is None: ticklen = 6    
    if xticklabels is None: xticklabels = True
    if yticklabels is None: yticklabels = True
    if ax is None:
        ax = plt.gca()
    else:
        plt.sca(ax)
    if xticks is None:
        xticks, xtext = plt.xticks()
    else:
        xtext = xticks
    if isinstance(xticklabels, Iterable):
        xtext = xticklabels
    elif xticklabels:
        xtext = list(xtext)
    else:
        xtext = len(xtext) * ['']
    if yticks is None:
        yticks, ytext = plt.yticks()
        if not any([str(i) for i in ytext]): ytext = yticks
    else:
        ytext = yticks
    if isinstance(yticklabels, Iterable):
        ytext = yticklabels
    elif yticklabels:
        ytext = list(ytext)
    else:
        ytext = len(ytext) * ['']
        
    xtext = list(xtext)
    ytext = list(ytext)
    if trim_to_limits:
        style_plot_limits(xticks, yticks)
    if not xtick0:
        xtext[0] = ''
    if not xtickf:
        xtext[-1] = ''
    if not ytick0:
        ytext[0] = ''
    if not ytickf:
        ytext[-1] = ''
    
    xticks = set_ticks(ax, xticks, 'x', xtext, xrot, xha, offset_xticks, fs, ticklen=ticklen)
    yticks = set_ticks(ax, yticks, 'y', ytext, yrot, yha, offset_yticks, fs, ticklen=ticklen)
    ax.zorder = 1
    xlim = plt.xlim()
    ylim = plt.ylim()
    axes = {'ax': ax}
    if right:
        x_twin = ax.twinx() if axtwinx is None else axtwinx
        ax._cached_xtwin = x_twin
        axes['twinx'] = x_twin 
        plt.sca(x_twin)
        x_twin.tick_params(axis='y', right=True, direction="in", length=ticklen)
        x_twin.zorder = 2
        plt.ylim(ylim)
        plt.yticks(yticks, ())
    if top:
        y_twin = ax.twiny() if axtwiny is None else axtwiny
        ax._cached_ytwin = y_twin
        axes['twiny'] = y_twin 
        plt.sca(y_twin)
        y_twin.tick_params(axis='x', top=True, direction="in", length=ticklen)
        y_twin.zorder = 2
        plt.xlim(xlim)
        plt.xticks(xticks, ())
    return axes
    
def style_plot_limits(xticks, yticks): # pragma: no cover
    plt.xlim([xticks[0], xticks[-1]])
    plt.ylim([yticks[0], yticks[-1]])
    
def fill_plot(color='k'): # pragma: no cover
    y_lb, y_ub = plt.ylim()
    plt.fill_between(plt.xlim(), [y_lb], [y_ub], color=color)
   
def set_axes_labels(axes, xlabel, ylabel): # pragma: no cover
    set_axes_xlabels(axes, xlabel)
    set_axes_ylabels(axes, ylabel)
   
def set_axes_xlabels(axes, xlabel): # pragma: no cover
    assert axes.ndim == 2
    N = axes.shape[1]
    if isinstance(xlabel, str): xlabel = N * [xlabel]
    for i in range(N): axes[-1, i].set_xlabel(xlabel[i])
    
def set_axes_ylabels(axes, ylabel): # pragma: no cover
    assert axes.ndim == 2
    N = axes.shape[0]
    if isinstance(ylabel, str): ylabel = N * [ylabel]
    for i in range(N): axes[i, 0].set_ylabel(ylabel[i])


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentVector/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import numpy as np
from thermosteam.base import (
    SparseVector, SparseLogicalVector, SparseArray, sparse_vector, sparse_array,
    nonzero_items, sparse
)
from numpy.testing import assert_allclose

def assert_no_zero_data(arr):
    if isinstance(arr, SparseVector):
        assert 0 not in arr.dct.values()
    elif isinstance(arr, SparseArray):
        assert all([0 not in i.dct.values() for i in arr if isinstance(i, SparseVector)])

def test_sparse_vector_creation():
    arr = np.array([1., 2., 0., 4.5])
    sv = sparse_vector(arr)
    assert_no_zero_data(sv)
    assert (sv.to_array() == arr).all()
    assert repr(sv) == 'sparse([1. , 2. , 0. , 4.5])'
    assert str(sv) == '[1.  2.  0.  4.5]'
    
    arr = np.array([1., 2., 0., 4.5, 0., 0.])
    sv = sparse_vector(arr)
    assert_no_zero_data(sv)
    assert (sv.to_array() == arr).all()
    assert repr(sv) == 'sparse([1. , 2. , 0. , 4.5, 0. , 0. ])'
    assert str(sv) == '[1.  2.  0.  4.5 0.  0. ]'
    assert (sparse(arr) == arr).all()
    
    sv_2 = sparse_vector(sv)
    assert sv_2.dct is sv.dct
    sv_2 = SparseVector(sv)
    assert sv_2.dct is not sv.dct
    
    sv = SparseVector(size=2)
    assert sv.size == 2
    assert (sv == 0).all()
    
    with pytest.raises(ValueError):
        SparseVector() # must pass size
        
    arr = np.array([True, True, False, True])
    sv = sparse_vector(arr)
    assert_no_zero_data(sv)
    assert (sv.to_array() == arr).all()
    assert repr(sv) == 'sparse([ True,  True, False,  True])'
    assert str(sv) == '[ True  True False  True]'
    assert (sparse(arr) == arr).all()
    
    sv_2 = sparse_vector(sv)
    assert sv_2.set is sv.set
    sv_2 = SparseLogicalVector(sv)
    assert sv_2.set is not sv.set
    assert sv.sparse_equal(sv_2)
    
    sv = SparseLogicalVector(SparseVector(size=2))
    arr = np.zeros(2, dtype=bool)
    assert (sv == arr).all()
    sv = SparseLogicalVector(size=2)
    assert sv.size == 2
    
    with pytest.raises(ValueError):
        SparseLogicalVector() # must pass size
    
def test_sparse_logical_vector_indexing_and_methods():
    arr = np.zeros(2, dtype=bool)
    sv = sparse(arr)
    assert (sv == 0).all()
    assert (sv[:1] == arr[:1]).all()
    assert (sv[:] == arr[:]).all()
    sv[0] = True
    arr[0] = True
    assert (sv == arr).all()
    sv[:] = arr[:] = [False, False]
    assert (sv == arr).all()
    sv[0] = True
    arr[0] = True
    sv[0] = False
    arr[0] = False
    assert (sv == arr).all()
    assert (~sv == ~arr).all()
    assert (abs(sv) == abs(arr)).all()
    sv[:] = True
    arr[:] = True
    assert (sv == arr).all()
    sv2 = sv.copy()
    sv2[:] = False
    sv[:] = sv2
    assert (sv2 == sv).all()
    
    sv[0] = -1 # This is casted to True
    assert not sv.has_negatives()
    assert not sv.has_negatives()
    for i in sv.nonzero_values(): assert i
    assert sv.sum_of(0)
    assert not sv.sum_of(1)
    assert sv.sum_of([0, 1])
    
    arr = np.ones(10, dtype=bool)
    sv = sparse(arr)
    sv[:2] = True
    arr[:2] = True
    assert (sv == arr).all()
    sv[0::2] = False
    arr[0::2] = False
    assert (sv == arr).all()
    sv[0:6:2] = True
    arr[0:6:2] = True
    assert (sv == arr).all()
    
    with pytest.raises(IndexError):
        sv[(1, 2)]
    with pytest.raises(IndexError):
        sv[(1, 2)] = 0.
    with pytest.raises(IndexError):
        sv[(1,)] = [0., 1]
    with pytest.raises(IndexError):
        sv[1] = [0., 1]
    
def test_sparse_logical_vector_miscellanious():
    for arr in [np.zeros(2, dtype=bool), np.ones(2, dtype=bool)]:
        sv = sparse(arr)
        for other in [True, [True], [True, True]]:
            sv /= other
            assert (sv == arr).all()
    
def test_sparse_array_creation():
    arr = np.array([[1., 2., 0., 4.5]])
    sa = sparse_array(arr)
    assert_no_zero_data(sa)
    assert (sa.to_array() == arr).all()
    assert repr(sa) == 'sparse([[1. , 2. , 0. , 4.5]])'
    assert str(sa) == '[[1.  2.  0.  4.5]]'
    
    arr = np.array([[1., 2., 0., 4.5, 0., 0.]])
    sa = sparse_array(arr)
    assert_no_zero_data(sa)
    assert (sa.to_array() == arr).all()
    assert repr(sa) == 'sparse([[1. , 2. , 0. , 4.5, 0. , 0. ]])'
    assert str(sa) == '[[1.  2.  0.  4.5 0.  0. ]]'
    
    assert (sparse(arr) == arr).all()
    
    arr = np.ones([2, 3, 1])
    with pytest.raises(ValueError):
        sparse(arr)
        
    sa = SparseArray()
    assert (sa == np.ones([0, 0])).all()
    
    with pytest.raises(TypeError):
        SparseArray(0)
        
    with pytest.raises(ValueError):
        bool(sa)
        
    arr = np.array([[True, True, False, True]])
    sa = sparse_array(arr)
    assert_no_zero_data(sa)
    assert (sa.to_array() == arr).all()
    assert repr(sa) == 'sparse([[ True,  True, False,  True]])'
    assert str(sa) == '[[ True  True False  True]]'
    
    arr = np.array([[True, True, False, True, False, False]])
    sa = sparse_array(arr)
    assert_no_zero_data(sa)
    assert (sa.to_array() == arr).all()
    assert repr(sa) == 'sparse([[ True,  True, False,  True, False, False]])'
    assert str(sa) == '[[ True  True False  True False False]]'
    assert (sparse(arr) == arr).all()
        
def test_sparse_vector_easy_methods():
    arr = np.array([1., 2., 0., 4.5])
    sv = sparse_vector(arr)
    assert [*sv.nonzero_items()] == nonzero_items(arr) == [(0, 1.), (1, 2.), (3, 4.5)]
    assert [*sv.nonzero_values()] == [1, 2, 4.5]
    nonzero_index, = sv.nonzero_index()
    assert [*sv.nonzero_keys()] == nonzero_index == [0, 1, 3]
    n_sv = -sv
    assert (n_sv == -arr).all()
    assert set(n_sv.negative_keys()) == n_sv.nonzero_keys()
    sv_index, = n_sv.negative_index()
    arr_index, = np.where(-arr < 0)
    (sv_index == arr_index).all()
    
    sv_index, = sv.positive_index()
    arr_index, = np.where(arr > 0)
    (sv_index == arr_index).all()
    
    n_sv.remove_negatives()
    assert (n_sv == np.zeros_like(sv)).all()

def test_sparse_array_easy_methods():
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse_array(arr)
    sa_left_index, sa_right_index = sa.nonzero_index()
    arr_left_index, arr_right_index = arr.nonzero()
    assert (sa_left_index == arr_left_index).all()
    assert (sa_right_index == arr_right_index).all()
    arr_nonzero_values = arr[arr_left_index, arr_right_index]
    assert (np.array([*sa.nonzero_values()]) == arr_nonzero_values).all()
    for ((i, j), k), ak in zip(sa.nonzero_items(), arr_nonzero_values):
        assert sa[i, j] == k == ak

    assert [*sa.negative_keys()] == []
    assert [*sa.negative_rows()] == []
    n_sa = -sa
    assert (n_sa == -arr).all()
    assert set(n_sa.negative_keys()) == sa.nonzero_keys()
    assert set(n_sa.negative_rows()) == {0, 1}
    n_sa.remove_negatives()
    assert (n_sa == np.zeros_like(sa)).all()
    
def test_sparse_vector_math():
    arr = np.array([1., -2., 0., 4.5])
    sv = sparse_vector(arr)
    arr = abs(arr)
    sv = abs(sv)
    assert (sv == arr).all()
    assert ((sv * 2) == (2 * arr)).all()
    assert ((sv + sv) == (2 * arr)).all()
    assert ((sv * 0) == ([0, 0, 0, 0])).all()
    assert ((sv - sv) == ([0, 0, 0, 0])).all()
    assert ((sv / sv) == (np.array([1., 1., 0., 1.]))).all()
    
    assert_no_zero_data(sv * 2)
    assert_no_zero_data(sv + sv)
    assert_no_zero_data(sv * 0)
    assert_no_zero_data(sv - sv)
    assert_no_zero_data(sv / sv)
    
    sv *= 0.
    assert sv.sparse_equal([0.])
    assert_no_zero_data(sv)
    
    sv += arr
    assert sv.sparse_equal(arr)
    assert_no_zero_data(sv)
    
    sv *= 2
    assert sv.sparse_equal(2. * arr)
    assert_no_zero_data(sv)
    
    sv /= 2.
    assert sv.sparse_equal(arr)
    assert_no_zero_data(sv)
    
    sv /= sv
    assert sv.sparse_equal(np.array([1., 1., 0., 1.]))
    assert_no_zero_data(sv)
    
    sv[:] *= 0.
    assert sv.sparse_equal([0.])
    assert_no_zero_data(sv)
    
    sv[:] += arr
    assert sv.sparse_equal(arr)
    assert_no_zero_data(sv)
    
    sv[:] *= 2
    assert sv.sparse_equal(2. * arr)
    assert_no_zero_data(sv)
    
    sv[:] /= 2.
    assert sv.sparse_equal(arr)
    assert_no_zero_data(sv)
    
    sv[:] /= sv
    assert sv.sparse_equal(np.array([1., 1., 0., 1.]))
    assert_no_zero_data(sv)
    sv[:] = 1
    assert_no_zero_data(sv)
    assert (2. / sv).sparse_equal([2., 2., 2., 2.])
    assert (sparse([2, 1, 0, 3]) / sv).sparse_equal([2, 1, 0, 3])
    
    sv[:] = [0., 1]
    assert_no_zero_data(sv)
    
    assert ((0 / sv) == np.zeros_like(sv)).all()
    
    with pytest.raises((ZeroDivisionError, FloatingPointError)):
        [2, 1, 0.1, 3] / sv
        
    with pytest.raises(ZeroDivisionError):
        sv /= 0

def test_sparse_logical_vector_math():
    left = [
        [True],
        [False],
        [0.],
        [3.],
        [-2.],
        [False, False],
        [True, True],
        [True, False, True],
        [[True, False, False], [False, True, False]],
        [[True]],
        [[False]],
        [0., 0.],
        [1., 3.],
        [1., 3., 0.],
        [[1., 2., 0.], [0., 0., 4.]],
        [[0.]],
        [[2.]],
    ]
    right = [
        [True],
        [False],
        [0.],
        [3.],
        [-2.],
        [False, False],
        [True, True],
        [True, True, False],
        [[True, True, False], [False, False, True]],
        [[[False]]],
        [[[True]]],
        [0., 0.],
        [1., 3.],
        [1., 3., 0.],
        [[1., 2., 0.], [0., 0., 4.]],
        [[[0.]]],
        [[[2.]]],
        0., 
        1.,
        3.,
    ]
    left = [sparse(i) for i in left]
    right = left + right + [np.array(i) for i in right]
    for L in left:
        for R in right:
            # Math
            try: value = np.asarray(L) + R
            except: pass
            else: assert ((L + R) == value).all()
            try: value = np.asarray(L) * R
            except: pass
            else: assert ((L * R) == value).all()
            try: value = np.asarray(L) / R
            except: pass
            else: assert ((L / R) == value).all()
            try: value = R / np.asarray(L)
            except: pass
            else: assert ((R / L) == value).all()
            try: value = np.asarray(L) - R
            except: pass
            else: assert ((L - R) == value).all()
            try: value = (np.asarray(L) & R)
            except: pass
            else: assert ((L & R) == value).all()
            try: value = (np.asarray(L) | R)
            except: pass
            else: assert ((L | R) == value).all()
            try: value = (np.asarray(L) ^ R)
            except: pass
            else: assert ((L ^ R) == value).all()
            
            # In-place math
            try: 
                arr = np.array(L)
                arr += R
            except: pass
            else: 
                s = L.copy()
                s += R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr *= R
            except: pass
            else: 
                s = L.copy()
                s *= R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr /= R
            except: pass
            else: 
                s = L.copy()
                s /= R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr -= R
            except: pass
            else: 
                s = L.copy()
                s -= R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr &= R
            except: pass
            else: 
                s = L.copy()
                s &= R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr |= R
            except: pass
            else: 
                s = L.copy()
                s |= R
                assert (s == arr).all()
            try: 
                arr = np.array(L)
                arr ^= R
            except: pass
            else: 
                s = L.copy()
                s ^= R
                assert (s == arr).all()
            
            # Equality math
            try: arr = (np.asarray(L) == R)
            except: pass
            else:
                if arr.__class__ is not bool: assert ((L == R) == arr).all()
            try: arr = (np.asarray(L) != R)
            except: pass
            else:
                if arr.__class__ is not bool: assert ((L != R) == arr).all()
            try: arr = (np.asarray(L) > R)
            except: pass
            else: 
                if arr.__class__ is not bool: assert ((L > R) == arr).all()
            try: arr = (np.asarray(L) < R)
            except: pass
            else:
                if arr.__class__ is not bool: assert ((L < R) == arr).all()
            try: arr = (np.asarray(L) >= R)
            except: pass
            else:
                if arr.__class__ is not bool: assert ((L >= R) == arr).all()
            try: arr = (np.asarray(L) <= R)
            except: pass
            else:
                if arr.__class__ is not bool: assert ((L <= R) == arr).all()
    
def test_sparse_array_math():
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse_array(-arr)
    assert_no_zero_data(sa)
    sa = abs(sa)
    assert (sa == arr).all()
    assert (sa * 2 == 2. * arr).all()
    assert (sa + sa == 2. * arr).all()
    assert (sa + arr == 2. * arr).all()
    assert (sa * sa == arr * arr).all()
    assert (sa + sa == 2. * arr).all()
    assert (sa * 0 == [[0.], [0.]]).all()
    assert (sa - sa == [[0.], [0.]]).all()
    assert (sa - 1 == arr - 1).all()
    assert (1 - sa == 1 - arr).all()
    assert (1 + sa == 1 + arr).all()
    assert (sa / sa == np.array([[1., 1., 0., 1.], [0., 0., 1., 1.]])).all()
    assert (sa / 1 == arr / 1).all()
    assert (sa / 2 == arr / 2).all()
    sparse_ones = sparse(np.ones([3, 2]))
    assert (2 / (2. * sparse_ones) == np.ones([3, 2])).all()
    assert (sparse_ones * 2 / sparse_ones == 2 * np.ones([3, 2])).all()
    
    assert_no_zero_data(sa * 2)
    assert_no_zero_data(sa + sa)
    assert_no_zero_data(sa * 0)
    assert_no_zero_data(sa - sa)
    assert_no_zero_data(sa / sa)
    sa[:] = 1.
    assert_allclose(arr.tolist() / sa, arr / sa.to_array())
    
    sa *= 0.
    assert sa.sparse_equal([[0.], [0.]])
    assert_no_zero_data(sa)
    
    sa += arr
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa -= arr
    assert sa.sparse_equal([[0.], [0.]])
    assert_no_zero_data(sa)
    
    sa += arr
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa *= 2
    assert sa.sparse_equal(2. * arr)
    assert_no_zero_data(sa)
    
    sa /= 2.
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa /= sa
    assert sa.sparse_equal(np.array([[1., 1., 0., 1.], [0., 0., 1., 1.]]))
    assert_no_zero_data(sa)
    
    sa[:] *= 0.
    assert sa.sparse_equal([[0.], [0.]])
    assert_no_zero_data(sa)
    
    sa[:] += arr
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa[:] -= arr
    assert sa.sparse_equal([[0.], [0.]])
    assert_no_zero_data(sa)
    
    sa[:] += arr
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa[:] *= 2
    assert sa.sparse_equal(2. * arr)
    assert_no_zero_data(sa)
    
    sa[:] /= 2.
    assert sa.sparse_equal(arr)
    assert_no_zero_data(sa)
    
    sa[:] /= sa
    assert sa.sparse_equal(np.array([[1., 1., 0., 1.], [0., 0., 1., 1.]]))
    assert_no_zero_data(sa)
    
    arr = np.array([1., -2., 0., 4.5])
    sv = sparse_vector(arr)
    assert (sv + [[2]] == arr + [[2]]).all()
    assert (sv - [[2]] == arr - [[2]]).all()
    assert (sv * [[2]] == arr * [[2]]).all()
    assert (sv / [[2]] == arr / [[2]]).all()
    assert ([[2]] / (sv + 1) == [[2]] / (arr + 1)).all()
    
    sa = sparse([[2]])
    assert (sv + sa == arr + sa).all()
    assert (sv - sa == arr - sa).all()
    assert (sv * sa == arr * sa).all()
    assert (sv / sa == arr / sa).all()
    assert (sa / (sv + 1) == sa / (arr + 1)).all()
    
def test_sparse_vector_indexing():
    arr = np.array([1., 2., 0., 4.5])
    sv = sparse_vector(arr)
    assert sv.size == 4 # Size is not strict
    old_dct = sv.dct.copy()
    sv[20] = 0.
    assert_no_zero_data(sv)
    assert sv.dct == old_dct
    sv[5] = 3.
    assert_no_zero_data(sv)
    assert sv.sparse_equal(np.array([1., 2., 0., 4.5, 0., 3.]))
    sv[[1, 3]] = [0., 0.]
    assert_no_zero_data(sv)
    assert sv.sparse_equal(np.array([1., 0., 0., 0., 0., 3.]))
    assert (sv[[0, 1, 5]] == np.array([1., 0., 3.])).all()
    assert sv[0] == 1.
    assert sv[4] == 0.
    assert sv[5] == 3.
    assert sv[10] == 0.
    sv[[0, 1, 5]] = 1.
    assert_no_zero_data(sv)
    assert sv.sparse_equal(np.array([1., 1., 0., 0., 0., 1.]))
    sv[:] = 2.
    assert_no_zero_data(sv)
    assert sv.size == 4
    assert sv.sparse_equal([2., 2., 2., 2.])
    sv[(1,)] = 3
    assert_no_zero_data(sv)
    assert sv[(1,)] == sv[1] == 3
    assert (sv[([1, 2],)] == np.array([3, 2])).all()
    assert (sv[[1, 2]] == np.array([3, 2])).all()
    
    arr = np.ones(10)
    sv = sparse(arr)
    sv[:2] = 3
    arr[:2] = 3
    assert (sv == arr).all()
    sv[0::2] = -2
    arr[0::2] = -2
    assert (sv == arr).all()
    sv[0:6:2] = -2
    arr[0:6:2] = -2
    assert (sv == arr).all()
    
    with pytest.raises(IndexError):
        sv[(1, 2)]
    with pytest.raises(IndexError):
        sv[(1, 2)] = 0.
    with pytest.raises(IndexError):
        sv[(1,)] = [0., 1]
    with pytest.raises(IndexError):
        sv[1] = [0., 1]

def test_sparse_array_indexing():
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse_array(arr)
    
    with pytest.raises(IndexError):
        sa[(1, 1, 0)]
    
    with pytest.raises(IndexError):
        sa[(1, 1, 0)] = 2
    
    old_sa = sa.copy()
    sa[0, 20] = 0.
    assert_no_zero_data(sa)
    assert (sa == old_sa).all()
    sa[0, 5] = 3.
    assert_no_zero_data(sa)
    assert sa[0].sparse_equal(np.array([1., 2., 0., 4.5, 0., 3.]))
    assert sa.sparse_equal(
        np.array([[1., 2., 0., 4.5, 0., 3.],
                  [0., 0., 1., 1.5, 0,  0.]])
    )
    assert sa[[0, 1]].sparse_equal(sa)
    sa[0, [0, 1]] = [0., 0.]
    assert_no_zero_data(sa)
    assert sa[0].sparse_equal(np.array([0. , 0. , 0. , 4.5, 0. , 3. ]))
    assert (sa[0, [0, 1, 5]] == np.array([0., 0., 3.])).all()
    assert sa[0, 0] == 0.
    assert sa[0, 4] == 0.
    assert sa[0, 5] == 3.
    assert sa[0, 10] == 0.
    sa[1] = 0
    assert_no_zero_data(sa)
    assert sa[1].sparse_equal([])
    assert sa[0].sparse_equal(np.array([0. , 0. , 0. , 4.5, 0. , 3. ]))
    sa[:, 1] = 2
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        np.array([[0., 2., 0., 4.5, 0., 3.],
                  [0., 2., 0., 0. , 0,  0.]])
    )
    sa[:, [1, 2]] = 1.
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        np.array([[0., 1., 1., 4.5, 0., 3.],
                  [0., 1., 1., 0. , 0,  0.]])
    )
    sa[:] = 0
    assert_no_zero_data(sa)
    assert sa.sparse_equal([])
    
    arr = np.array([1., 2., 0., 4.5])
    sv = sparse_vector(arr)
    sa[0, :] = sv
    assert_no_zero_data(sa)
    assert sa[0, :].sparse_equal(arr)
    sa[0, :] = 0.
    assert_no_zero_data(sa)
    assert sa[0, :].sparse_equal([])
    sa[[0, 1], :] = sv
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[1., 2., 0., 4.5],
         [1., 2., 0., 4.5]]
    )
    sa[[0, 1], :] = [0, 1, 0]
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[0, 1],
         [0, 1]]
    )
    sa[[0, 1], :] = [[0, 1], [1, 0]]
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[0, 1],
         [1, 0]]
    )
    sa[:] = 0
    assert_no_zero_data(sa)
    sa[:, [1, 3]] = [[0, 1],
                     [1, 0]]
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[0, 0, 0, 1],
         [0, 1, 0, 0]]
    )
    sa[:] = 2.
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[2., 2., 2., 2.],
         [2., 2., 2., 2.]]
    )
    sa[[0, 1], :] = 3.
    assert_no_zero_data(sa)
    assert sa.sparse_equal(
        [[3., 3., 3., 3.],
         [3., 3., 3., 3.]]
    )
    with pytest.raises(IndexError):
        sa[[[0, 1], [2, 3]]] = 2
    
    sa[[0, 1], [2, 3]] = 2
    assert_no_zero_data(sa)
    assert (sa[[0, 1], [2, 3]] == np.array([2., 2.])).all()
    assert sa.sparse_equal(
        [[3., 3., 2., 3.],
         [3., 3., 3., 2.]]
    )

    sa_not_a_view = sa[:, [1, 2]]
    with pytest.raises(ValueError):
        sa_not_a_view[:] = 20
    
    assert (sa_not_a_view ==
        [[3., 2.],
         [3., 3.]]
    ).all()
    
    arr = sa.to_array()
    assert (sa[[0, 1], :] == arr[[0, 1], :]).all()
    
    with pytest.raises(IndexError):
        sa[[[0, 1]], :]
    
    with pytest.raises(IndexError):
        sa[[[0, 1]], [[0]]]    
    
    assert (sa[[0, 1], 0] == arr[[0, 1], 0]).all()
    
    with pytest.raises(IndexError):
        sa[[[1, 2], 3]]
    
    sa[:, [0, 1]] = [5, 6]
    arr[:, [0, 1]] = [5, 6]
    assert (sa[:, [0, 1]] == [5, 6]).all()
    assert (arr[:, [0, 1]] == [5, 6]).all()

    sa[[0, 1], :] = 0
    arr[[0, 1], :] = 0
    assert (arr[[0, 1], :] == sa[[0, 1], :]).all()
    
    sa[0:2, :] = 2
    arr[0:2, :] = 2
    assert (sa[0:2, :] == arr[0:2, :]).all()
    sa[:, :1] = 2
    arr[:, :1] = 2
    assert (sa[:, :1] == arr[:, :1]).all()
    sa[:2] = 2 
    arr[:2] = 2
    assert (sa[:3] == arr[:3]).all()
    
    arr = np.ones([3,4])
    sa = sparse(arr)
    sa[[True, False, True]] = 2
    arr[[True, False, True]] = 2
    assert (sa == arr).all() and (sa[[True, False, True]] == arr[[True, False, True]]).all()
    sa[[True, False, True], :2] = 3
    arr[[True, False, True], :2] = 3
    assert (sa == arr).all() and (sa[[True, False, True], :2] == arr[[True, False, True], :2]).all()
    sa[0, :2] = 3
    arr[0, :2] = 3
    assert (sa == arr).all() and (sa[0, :2] == arr[0, :2]).all()
    sa[0, 2:4] = 10
    arr[0, 2:4] = 10
    assert (sa == arr).all() and (sa[0, 2:4] == arr[0, 2:4]).all()
    sa[[0, 2]] = 0
    arr[[0, 2]] = 0
    assert (sa == arr).all() and (sa[[0, 2]] == arr[[0, 2]]).all()
    sa[[0, 1], [2, 0]] = 0
    arr[[0, 1], [2, 0]] = 0
    assert (sa == arr).all() and (sa[[0, 1], [2, 0]] == arr[[0, 1], [2, 0]]).all()

    with pytest.raises(IndexError):
        sa[(0,)] # Too few indices, array is 2 dimensional but only 1 dimension was indexed
        
    assert (sa[[1], :2] == arr[[1], :2]).all()
    
    with pytest.raises(IndexError):
        sa[[[0]], :] # Row index dimensions can be at most 1
    
    with pytest.raises(IndexError):
        sa[[0, 1], [[0]]] # Column index dimensions can be at most 1
        
    sa[[0, 1]] = [[[0, 1]]] # Broadcast 3-d array to 2-d
        
    sa[[0, 1], 0] = [[0, 1]] # Reduce dimensions
    
    # Check closed and open slices
    sa = SparseArray.from_shape([5, 5])
    sa[0:3, :] = 1
    arr = np.zeros([5, 5])
    arr[0:3, :] = 1
    assert (sa == arr).all()
    
    sa = SparseArray.from_shape([5, 5])
    sa[:, :] = 1
    arr = np.zeros([5, 5])
    arr[:, :] = 1
    assert (sa == arr).all()

def test_sparse_array_boolean_indexing():
    for dtype in (float, bool):
        arr = np.array([[1., 0.], [0., 1.]], dtype=dtype)
        sa = sparse_array(arr)
        assert (sa[sa == 0] == arr[arr == 0]).all()
        sa[sa == 0] = 1.
        arr[arr == 0] = 1.
        assert (sa == arr).all()
        sa[sa == 1] = [0., 1, 1, 0.]
        arr[arr == 1] = [0., 1, 1, 0.]
        assert (sa == arr).all()
        mask = (arr == 0).any(0)
        assert (sa[mask] == arr[mask]).all()
        if dtype is float:
            arr[mask] *= 2
            sa[mask] *= 2
            assert (sa[mask] == arr[mask]).all()
            sa[mask] = 2
            arr[mask] = 2
            assert (sa[mask] == arr[mask]).all()
        elif dtype is bool:
            arr[mask] &= True
            sa[mask] &= True
            assert (sa[mask] == arr[mask]).all()
            sa[mask] = False
            arr[mask] = False
            assert (sa[mask] == arr[mask]).all()
        arr = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]], dtype=dtype)
        sa = SparseArray(arr)
        assert (arr[[True, False, True], :] == sa[[True, False, True], :]).all()
        assert (arr[:, [True, False, True]] == sa[:, [True, False, True]]).all()

def test_sparse_vector_boolean_indexing():
    for dtype in (float, bool):
        arr = np.array([1., 0., 0., 1.], dtype=dtype)
        sv = sparse_vector(arr)
        assert (sv[sv == 0] == arr[arr == 0]).all()
        sv[sv == 0] = 1.
        arr[arr == 0] = 1.
        assert (sv == arr).all()
        sv[sv == 1] = [0., 1, 1, 0.]
        arr[arr == 1] = [0., 1, 1, 0.]
        assert (sv == arr).all()

def test_sparse_vector_special_methods():
    for dtype in (float, bool):
        arr = np.array([1., 2., 0., 4.5], dtype=dtype)
        sv = sparse_vector(arr)
        assert (sv.to_flat_array() == arr).all()
        arr = np.array([1., 2., 0., 2], dtype=dtype)
        sv.from_flat_array(arr)
        assert (sv.to_flat_array() == arr).all()
        
        with pytest.raises(IndexError):
            sv[[[True, False]]] # boolean index dimensions is off
        
        with pytest.raises(IndexError):
            sv[[[True, False]]] = [2, 2] # boolean index dimensions is off
        
        with pytest.raises(IndexError):
            sv[[[1, 2]]] # index dimensions is off
    
        with pytest.raises(IndexError):
            sv[[[1, 2]]] = [2, 3] # index dimensions is off
            
        sv[[1, 2]] = [[2, 3]] # Reduce dimensions
        
        sv[:] = [[2]] # Reduce dimensions

def test_sparse_array_special_methods():
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse_array(arr)
    assert (sa.to_flat_array() == arr.flatten()).all()
    arr = np.array([1., 2., 0., 0, 0., 0., 1., 2])
    sa.from_flat_array(arr)
    assert (sa.to_flat_array() == arr.flatten()).all()
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse(arr)
    assert (sa.sum_of([3, 1], axis=0) == [6., 2.]).all()
    assert (sa.sum_of([0, 2], axis=1) == [1, 1]).all()
    with pytest.raises(ValueError):
        sa.sum_of([0, 2], axis=2) # Invalid axis
    assert set(zip(*sa.positive_index())) == set([(int(i), int(j)) for i,j in zip(*np.where(arr > 0))])
    assert set(zip(*(-sa).negative_index())) == set([(int(i), int(j)) for i,j in zip(*np.where(arr > 0))])
    assert sa.nonzero_rows() == [0, 1]
    sa[:] = 0
    assert sa.nonzero_rows() == []
    assert [*nonzero_items(sa)] == [*sa.nonzero_items()] == [*nonzero_items(sa.to_array())]
    assert [*nonzero_items(sa[0])] == [*sa[0].nonzero_items()]
    
    sa = sparse([[1, 2], [3, 4]])
    sa_other = sa.copy()
    sa.copy_like(sparse(sa + 1))
    assert (sa == sparse(sa_other + 1)).all()
    
    # Now with booleans
    
    arr = np.array([[True, True, False, True], [True, True, False, True]])
    sa = sparse_array(arr)
    assert (sa.to_flat_array() == np.array([True, True, False, True, True, True, False, True])).all()
    sa.from_flat_array(np.array([True, True, False, True, True, True, False, True]))
    assert (sa == np.array([[True, True, False, True], [True, True, False, True]])).all()
    
    arr = np.array([[True, True, False, True], [False, False, True, True]])
    sa = sparse_array(arr)
    assert (sa.to_flat_array() == arr.flatten()).all()
    arr = np.array([True, True, False, False, False, False, True, True])
    sa.from_flat_array(arr)
    assert (sa.to_flat_array() == arr.flatten()).all()
    arr = np.array([[True, True, False, True], [False, False, True, True]])
    sa = sparse(arr)
    assert (sa.sum_of([3, 1], axis=0) == [2, 1]).all()
    assert (sa.sum_of([0, 2], axis=1) == [1, 1]).all()
    with pytest.raises(ValueError):
        sa.sum_of([0, 2], axis=2) # Invalid axis
    assert set(zip(*sa.positive_index())) == set([(int(i), int(j)) for i,j in zip(*np.where(arr > 0))])
    assert set(zip(*(-sa).negative_index())) == set([(int(i), int(j)) for i,j in zip(*np.where(arr > 0))])
    assert sa.nonzero_rows() == [0, 1]
    sa[:] = 0
    assert sa.nonzero_rows() == []
    assert [*nonzero_items(sa)] == [*sa.nonzero_items()] == [*nonzero_items(sa.to_array())]
    assert [*nonzero_items(sa[0])] == [*sa[0].nonzero_items()]
    
    sa = sparse([[1, 2], [3, 4]])
    sa_other = sa.copy()
    sa.copy_like(sparse(sa + 1))
    assert (sa == sparse(sa_other + 1)).all()
    
    arr = np.array([[True, True, False, True], [True, True, False, True]])
    sa = sparse_array(arr)
    assert (sa.to_flat_array() == np.array([True, True, False, True, True, True, False, True])).all()
    sa.from_flat_array(np.array([True, True, False, True, True, True, False, True]))
    assert (sa == np.array([[True, True, False, True], [True, True, False, True]])).all()
    
def test_descriptive_methods():
    for arr in [np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]]),
                np.array([[True, True, False, True], [False, False, True, True]])]:
        sa = sparse(arr)
        assert (sa.value == sa.to_array()).all()
        assert sa.size == arr.size
        assert sa.vector_size == sa[0].size == sa[0].vector_size == arr[0].size
        assert sa.shape == arr.shape
        assert sa[0].shape == arr[0].shape

def test_read_only_flag():
    arr = np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]])
    sa = sparse(arr)
    sa.setflags(0)
    with pytest.raises(ValueError):
        sa[:] = 1
   
    with pytest.raises(NotImplementedError):
        sa.setflags(1)
   
    arr = np.ones(3)
    sv = sparse(arr)
    sv.read_only = True
    with pytest.raises(ValueError):
        sv[:] = 1
        
    sv.read_only = False
    sv.setflags(0)
    assert sv.read_only
    with pytest.raises(NotImplementedError):
        sv.setflags(1) # flag not implemented

def test_sparse_vector_methods_vs_numpy():
    for dtype in (float, bool):
        arrays = [
            np.array([1., 2., 0., 4.5], dtype=dtype),
            np.zeros(3, dtype=dtype),
            np.ones(3, dtype=dtype),
        ]
        for arr in arrays:
            sv = sparse_vector(arr)
            for method in ('min', 'max', 'argmin', 'argmax', 'mean', 'sum', 'any', 'all'):
                sv_method = getattr(sv, method)
                arr_method = getattr(arr, method)
                for axis in (0, None, 1):
                    for keepdims in (False, True):
                        if axis == 1:
                            with pytest.raises(ValueError):
                                np.asarray(sv_method(axis=axis, keepdims=keepdims))
                            with pytest.raises(ValueError): # For good measure
                                arr_method(axis=axis, keepdims=keepdims)
                            continue
                        sv_result = np.asarray(sv_method(axis=axis, keepdims=keepdims))
                        arr_result = arr_method(axis=axis, keepdims=keepdims)
                        assert sv_result.shape == arr_result.shape, f"wrong shape in SparseVector.{method} with axis {axis}, dtype {dtype}, and keepdims {keepdims}"
                        assert (sv_result == arr_result).all(), f"wrong value in SparseVector.{method} with axis {axis}, dtype {dtype}, and keepdims {keepdims}"

def test_sparse_array_methods_vs_numpy():
    for dtype in (float, bool):
        arrays = [
            np.array([[1., 2., 0., 4.5], [0., 0., 1., 1.5]], dtype=dtype),
            np.zeros([2, 3], dtype=dtype),
            np.ones([2, 3], dtype=dtype),
        ]
        for arr in arrays:
            sa = sparse_array(arr)
            for method in ('min', 'max', 'argmin', 'argmax', 'mean', 'sum', 'any', 'all'):
                sa_method = getattr(sa, method)
                arr_method = getattr(arr, method)
                for axis in (0, 1, None, 2):
                    for keepdims in (False, True):
                        if axis == 2:
                            with pytest.raises(ValueError):
                                np.asarray(sa_method(axis=axis, keepdims=keepdims))
                            with pytest.raises(ValueError): # For good measure
                                arr_method(axis=axis, keepdims=keepdims)
                            continue
                        sa_result = np.asarray(sa_method(axis=axis, keepdims=keepdims))
                        arr_result = arr_method(axis=axis, keepdims=keepdims)
                        assert sa_result.shape == arr_result.shape, f"wrong shape in SparseArray.{method} with axis {axis}, dtype {dtype}, and keepdims {keepdims}"
                        assert (sa_result == arr_result).all(), f"wrong value in SparseArray.{method} with axis {axis}, dtype {dtype}, and keepdims {keepdims}"
        sa = sparse([[]])
        assert sa.vector_size == 0
        with pytest.raises(ValueError):
            sa.min()
        with pytest.raises(ValueError):
            sa.max()

if __name__ == '__main__':
    test_sparse_vector_creation()
    test_sparse_array_creation()
    test_sparse_vector_easy_methods()
    test_sparse_array_easy_methods()
    test_sparse_vector_math()
    test_sparse_logical_vector_math()
    test_sparse_logical_vector_miscellanious()
    test_sparse_logical_vector_indexing_and_methods()
    test_sparse_array_math()
    test_sparse_vector_indexing()
    test_sparse_array_indexing()
    test_sparse_vector_boolean_indexing()
    test_sparse_array_boolean_indexing()
    test_sparse_vector_special_methods()
    test_sparse_array_special_methods()
    test_descriptive_methods()
    test_read_only_flag()
    test_sparse_vector_methods_vs_numpy()
    test_sparse_array_methods_vs_numpy()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
from setuptools import setup

setup(
    name='thermosteam',
    packages=['thermosteam'],
    license='MIT',
    version='0.52.16',
    description="BioSTEAM's Premier Thermodynamic Engine",
    long_description=open('README.rst', encoding='utf-8').read(),
    author='Yoel Cortes-Pena',
    install_requires=['numpy==1.26.4',
                      'pint==0.24.4',
                      'scipy>=1.15.2',
                      'thermo>=0.6.0',
                      'IPython>=7.9.0',
                      'colorpalette>=0.3.3',
                      'pandas>=2.2.2',
                      'matplotlib>=3.10.1',
                      'xlrd>=1.2.0',
                      'openpyxl>=3.0.0',
                      'flexsolve>=0.5.9',
                      'numba==0.62.1',
                      'coolprop',
                      'imageio',
                      'pydot',
                      'pyglet'],
    extras_require={
        'dev': [
            'biorefineries>=2.23.16',
            'sympy',
            'sphinx',
            'sphinx_rtd_theme',
            'pyyaml',
            'pytest-cov',
            'coveralls',
        ]
    },
    package_data={
        'thermosteam': [
            'temporary_images',
            'base/*',
            'equilibrium/*',
            'reaction/*',
            'utils/*',
            'utils/decorators/*',
            'mixture/*',
            'chemicals/*',
            'chemicals/Reaction/*',
            'equilibrium/UNIFAC/*',
            'thermo/*',
            'units_of_measure.txt',
        ]
    },
    excluded=['*.png'],
    python_requires='>=3.9',
    platforms=['Windows', 'Mac', 'Linux'],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/BioSTEAMDevelopmentGroup/thermosteam',
    download_url='https://github.com/BioSTEAMDevelopmentGroup/thermosteam.git',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'License :: OSI Approved :: University of Illinois/NCSA Open Source License',
                 'License :: OSI Approved :: MIT License',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Manufacturing',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English',
                 'Operating System :: MacOS',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Unix',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Education'],
    keywords=['thermodynamics', 'chemical engineering', 'mass and energy balance', 'material properties', 'phase equilibrium'],
)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ..base import display_asfunctor, SparseVector
from math import log

__all__ = (
    'IdealTMixtureModel', 'IdealTPMixtureModel', 'IdealEntropyModel',
    'IdealHvapModel', 'SinglePhaseIdealTMixtureModel', 'SinglePhaseIdealTPMixtureModel',
)

class IdealTPMixtureModel:
    """
    Create an IdealTPMixtureModel object that calculates mixture properties
    based on the molar weighted sum of pure chemical properties.
    
    Parameters
    ----------
    models : Iterable[function(T, P)]
        Chemical property functions of temperature and pressure.
    var : str
        Description of thermodynamic variable returned.
    
    Notes
    -----
    :class:`Mixture` objects can contain IdealMixtureModel objects to establish
    as mixture model for thermodynamic properties.
    
    See also
    --------
    :class:`Mixture`
    :func:`~.mixture_builders.ideal_mixture`
    
    Examples
    --------
    >>> from thermosteam.mixture import IdealTPMixtureModel
    >>> from thermosteam import Chemicals
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> models = [i.V for i in chemicals]
    >>> mixture_model = IdealTPMixtureModel(models, 'V')
    >>> mixture_model
    <IdealTPMixtureModel(phase, mol, T, P) -> V [m^3/mol]>
    >>> mixture_model('l', [0.2, 0.8], 350, 101325)
    5.364...-05
    
    """
    __slots__ = ('var', 'models')

    def __init__(self, models, var):
        self.models = tuple(models)
        self.var = var

    def __call__(self, phase, mol, T, P):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        models = self.models
        return sum([j * models[i](phase, T, P) for i, j in mol.dct.items()])
    
    def __repr__(self):
        return f"<{display_asfunctor(self)}>"

class IdealEntropyModel:
    """
    Create an IdealEntropyModel object that calculates entropy of a mixture 
    based on the molar weighted sum of pure chemical properties.
    
    Parameters
    ----------
    models : Iterable[function(T, P)]
        Chemical property functions of temperature and pressure.
    var : str
        Description of thermodynamic variable returned.
    
    Notes
    -----
    :class:`Mixture` objects can contain IdealMixtureModel objects to establish
    as mixture model for thermodynamic properties.
    
    See also
    --------
    :class:`Mixture`
    :func:`~.mixture_builders.ideal_mixture`
    
    Examples
    --------
    >>> from thermosteam.mixture import IdealEntropyModel
    >>> from thermosteam import Chemicals
    >>> import numpy as np
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> models = [i.S for i in chemicals]
    >>> mixture_model = IdealEntropyModel(models, 'S')
    >>> mixture_model
    <IdealEntropyModel(phase, mol, T, P) -> S [J/K/mol]>
    >>> mixture_model('l', np.array([0.2, 0.8]), 350, 101325)
    160.174
    
    """
    __slots__ = IdealTPMixtureModel.__slots__
    __init__ = IdealTPMixtureModel.__init__
    __repr__ = IdealTPMixtureModel.__repr__

    def __call__(self, phase, mol, T, P):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        total_mol = mol.sum()
        models = self.models
        return sum([j * models[i](phase, T, P) + j * log(j / total_mol) for i, j in mol.dct.items()])
    

class IdealTMixtureModel:
    """
    Create an IdealTMixtureModel object that calculates mixture properties
    based on the molar weighted sum of pure chemical properties.
    
    Parameters
    ----------
    models : Iterable[function(T, P)]
        Chemical property functions of temperature and pressure.
    var : str
        Description of thermodynamic variable returned.
    
    Notes
    -----
    :class:`Mixture` objects can contain IdealMixtureModel objects to establish
    as mixture model for thermodynamic properties.
    
    See also
    --------
    :class:`Mixture`
    
    Examples
    --------
    >>> from thermosteam.mixture import IdealTMixtureModel
    >>> from thermosteam import Chemicals
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> models = [i.Cn for i in chemicals]
    >>> mixture_model = IdealTMixtureModel(models, 'Cn')
    >>> mixture_model
    <IdealTMixtureModel(phase, mol, T, P=None) -> Cn [J/mol/K]>
    >>> mixture_model('l', [0.2, 0.8], 350)
    122.529
    
    """
    __slots__ = IdealTPMixtureModel.__slots__
    __init__ = IdealTPMixtureModel.__init__
    __repr__ = IdealTPMixtureModel.__repr__

    def __call__(self, phase, mol, T, P=None):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        models = self.models
        return sum([j * models[i](phase, T) for i, j in mol.dct.items()])


class SinglePhaseIdealTMixtureModel:
    """
    Create an SinglePhaseIdealTMixtureModel object that calculates mixture properties
    based on the molar weighted sum of pure chemical properties.
    
    Parameters
    ----------
    models : Iterable[function(T, P)]
        Chemical property functions of temperature and pressure.
    var : str
        Description of thermodynamic variable returned.
    
    Notes
    -----
    :class:`Mixture` objects can contain IdealMixtureModel objects to establish
    as mixture model for thermodynamic properties.
    
    See also
    --------
    :class:`Mixture`
    
    Examples
    --------
    >>> from thermosteam.mixture import SinglePhaseIdealTMixtureModel
    >>> from thermosteam import Chemicals
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> models = [i.Psat for i in chemicals]
    >>> mixture_model = SinglePhaseIdealTMixtureModel(models, 'Psat')
    >>> mixture_model
    <SinglePhaseIdealTMixtureModel(mol, T, P=None) -> Psat [Pa]>
    >>> mixture_model([0.2, 0.8], 350)
    84499.12
    
    """
    __slots__ = IdealTPMixtureModel.__slots__
    __init__ = IdealTPMixtureModel.__init__
    __repr__ = IdealTPMixtureModel.__repr__

    def __call__(self, mol, T, P=None):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        models = self.models
        return sum([j * models[i](T) for i, j in mol.dct.items()])

class SinglePhaseIdealTPMixtureModel:
    """
    Create an IdealTPMixtureModel object that calculates mixture properties
    based on the molar weighted sum of pure chemical properties.
    
    Parameters
    ----------
    models : Iterable[function(T, P)]
        Chemical property functions of temperature and pressure.
    var : str
        Description of thermodynamic variable returned.
    
    Notes
    -----
    :class:`Mixture` objects can contain IdealMixtureModel objects to establish
    as mixture model for thermodynamic properties.
    
    See also
    --------
    :class:`Mixture`
    :func:`~.mixture_builders.ideal_mixture`
    
    Examples
    --------
    >>> from thermosteam.mixture.ideal_mixture_model import SinglePhaseIdealTPMixtureModel
    >>> from thermosteam import Chemicals
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> models = [i.V.l for i in chemicals]
    >>> mixture_model = SinglePhaseIdealTPMixtureModel(models, 'V')
    >>> mixture_model
    <SinglePhaseIdealTPMixtureModel(mol, T, P) -> V [m^3/mol]>
    >>> mixture_model([0.2, 0.8], 350, 101325)
    5.3640...-05
    
    """
    __slots__ = ('var', 'models')

    def __init__(self, models, var):
        self.models = tuple(models)
        self.var = var

    def __call__(self, mol, T, P):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        models = self.models
        return sum([j * models[i](T, P) for i, j in mol.dct.items()])
    
    def __repr__(self):
        return f"<{display_asfunctor(self)}>"

class IdealHvapModel:
    __slots__ = ('chemicals',)
    var = 'Hvap'

    def __init__(self, chemicals):
        self.chemicals = chemicals

    def __call__(self, mol, T, P=None):
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        chemicals = self.chemicals
        return sum([
            j * k.Hvap(T) for i, j in mol.dct.items()
            if not (k:=chemicals[i]).locked_state
        ])
    
    __repr__ = IdealTPMixtureModel.__repr__



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the t_dependent_property module from the thermo library:
# https://github.com/CalebBell/thermo
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from thermo import TDependentProperty, VolumeSolid
from warnings import warn
TDependentProperty.RAISE_PROPERTY_CALCULATION_ERROR = True

# Remove cache from call
TDependentProperty.__call__ = TDependentProperty.T_dependent_property

### Backwards compatibility with past thermosteam versions

# Allow for decorators
def add_model(self, f=None, *args, top_priority=True, **kwargs):
    if f is None:
        return lambda f: self.add_method(f, *args, **kwargs)
    else:
        self.add_method(f, *args, **kwargs)

TDependentProperty.add_model = add_model

# Missing method 
def has_method(self):
    return bool(self._method)

TDependentProperty.__bool__ = has_method

# Consistency with other volume phases
def __call__(self, T, P=None):
    return self.T_dependent_property(T)

VolumeSolid.__call__ = __call__

# Shallow copy
def copy(self):
    cls = type(self)
    copy = cls.__new__(cls)
    copy.__dict__.update(self.__dict__)
    return copy

TDependentProperty.copy = copy

# Handling methods

def _set_method(self, method, stacklevel):
    if method is not None:
        old_method = method
        if method not in self.all_methods and method != 'POLY_FIT':
            old_method = method
            method, *_ = method.split('(')
            method = method.upper().replace(' ', '_').replace('_AND_', '_').strip('_').replace('SOLID', 'S').replace('SATURATION', 'SAT')
        if method not in self.all_methods and method != 'POLY_FIT':
            raise ValueError("Method '%s' is not available for this chemical; "
                             "available methods are %s" %(method, self.all_methods))
        if old_method != method:
            warn("Method '%s' does not exist, but was assummed to indicate '%s'" %(old_method, method),
                 category=RuntimeWarning, stacklevel=stacklevel)
    self._method = method

@TDependentProperty.method.setter
def method(self, method):
    self._set_method(method, 2)

def set_model_priority(self, model, priority=0):
    if priority == 0:
        warn("'set_model_priority' is deprecated; set the 'method' "
             "attribute to change model", DeprecationWarning)
        self._set_method(model, 3)
    else:
        raise RuntimeError(
            "'set_model_priority' is deprecated; cannot set model "
            "priority as models are not cycled anymore")
        
def move_up_model_priority(self, model, priority=0):
    if priority == 0:
        warn("'move_up_model_priority' is deprecated; set the 'method' "
             "attribute to change model", DeprecationWarning)
        self._set_method(model, 3)
    else:
        raise RuntimeError(
            "'move_up_model_priority' is deprecated; cannot set model "
            "priority as models are not cycled anymore")

TDependentProperty._set_method = _set_method
TDependentProperty.method = method
TDependentProperty.set_model_priority = set_model_priority
TDependentProperty.move_up_model_priority = move_up_model_priority


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import thermosteam as tmo
from numpy.testing import assert_allclose

def test_aliases():
    chemicals = tmo.Chemicals(['Water', 'Ethanol', 'Methanol'], cache=True)
    chemicals.compile()
    chemicals.define_group('alcohols', ['Water', 'Ethanol'])
    assert set(chemicals.get_aliases('H2O')) == {'7732-18-5', 'H2O', 'WATER', 'Water', 'oxidane', 'water'}
    
def test_chemical_cache():
    
    @tmo.utils.chemical_cache
    def create_ethanol_chemicals(other_chemicals):
        chemicals = tmo.Chemicals(['Water', 'Ethanol', *other_chemicals])
        return chemicals
    
    chemicals_a = create_ethanol_chemicals(('CO2', 'O2'))
    chemicals_b = create_ethanol_chemicals(('CO2', 'O2'))
    assert chemicals_a is chemicals_b
    
    chemicals_c = create_ethanol_chemicals(('O2',))
    assert chemicals_a is not chemicals_c 
    assert chemicals_a.Water is chemicals_c.Water
    assert chemicals_b.O2 is chemicals_c.O2
    
    
if __name__ == '__main__':
    test_aliases()
    test_chemical_cache()